"""Low-level SLURM API to submit, cancel and monitor jobs"""

import os
import time
import json
import logging
import datetime
from pprint import pformat
from urllib.parse import urlparse, urlunparse
from contextlib import contextmanager
from typing import List, Optional, Tuple, Union, Generator, Dict, Any

from .. import utils
from .. import defaults
from .. import os_utils
from .. import url_utils
from .api import slurm_logs
from .api import slurm_access
from .api import slurm_params
from .api import slurm_response
from .api import slurm_auth
from ..log_utils import log_file_monitor_context
from ..errors import SlurmInvalidUrlError
from ..errors import SlurmMissingParameterError

from pydantic import BaseModel

logger = logging.getLogger(__name__)


class SlurmBaseRestClient:
    """Low-level SLURM API to submit, cancel and monitor jobs.
    This class does not contain any job-related state."""

    def __init__(
        self,
        url: str = "",
        user_name: str = "",
        token: str = "",
        api_version: str = "",
        renewal_url: str = "",
        parameters: Optional[dict] = None,
        log_directory: Optional[str] = None,
        std_split: Optional[bool] = False,
        request_options: Optional[dict] = None,
        use_os_environment: bool = True,
    ):
        """
        :param url: SLURM REST API URL (fallback to SLURM_URL env)
        :param user_name: SLURM username (fallback to SLURM_USER or system user)
        :param token: SLURM JWT token (fallback to SLURM_TOKEN env)
        :param api_version: SLURM API version (e.g. 'v0.0.41')
        :param renewal_url: Url for SLURM JWT token renewal (fallback to SLURM_RENEWAL_URL env)
        :param parameters: SLURM job parameters
        :param log_directory: SLURM log directory
        :param std_split: Split standard output and standard error
        :param request_options: GET, POST and DELETE options
        :param use_os_environment: Use ``SLURM_*`` environment variables
        :raises SlurmMissingParameterError:
        :raises SlurmTokenInvalidError:
        :raises SlurmInvalidUrlError:
        """
        if not url and use_os_environment:
            url = os.environ.get("SLURM_URL")
        if not url:
            raise SlurmMissingParameterError(
                "'url' must be provided or set environment variable SLURM_URL"
            )
        self._base_url = _parse_slurm_rest_url(url)

        self._auth = slurm_auth.SlurmAuthManager(
            user_name=user_name,
            token=token,
            renewal_url=renewal_url,
            use_os_environment=use_os_environment,
        )

        if not api_version and use_os_environment:
            api_version = os.environ.get("SLURM_API_VERSION")
        if not api_version:
            api_version = defaults.DEFAULT_API_VERSION
        version_tuple = slurm_access.parse_version(api_version)
        self._api_version_str: str = slurm_access.create_version(version_tuple)
        self._api_version_tuple: Tuple[int, int, int] = version_tuple

        self._parameters = parameters
        if log_directory:
            log_directory = str(log_directory).format(user_name=self._auth.user_name)
        self._log_directory = log_directory
        self._std_split = std_split
        self._request_options = request_options

    def submit_job(
        self,
        script: str,
        parameters: Optional[dict] = None,
        metadata: Optional[Union[str, dict]] = None,
        request_options: Optional[dict] = None,
    ) -> int:
        """Returns the SLURM job ID"""
        # Job parameters
        parameters = utils.merge_mappings(self._parameters, parameters)
        self._submit_ensure_name(parameters)
        self._submit_ensure_std(parameters)
        self._submit_ensure_environment(parameters)
        self._submit_ensure_wd(parameters)
        if metadata:
            if not isinstance(metadata, str):
                metadata = json.dumps(metadata)
            parameters["comment"] = metadata
        slurm_params.coerce_parameters(self._api_version_tuple, parameters)

        # Request body
        body = {"job": parameters}
        if request_options:
            request_options["json"] = body
        else:
            request_options = {"json": body}

        # Job script
        if self._api_version_tuple < (0, 0, 39):
            body["script"] = script
        else:
            body["job"]["script"] = script

        # Submit job
        response = self.post(
            f"/slurm/{self._api_version_str}/job/submit",
            request_options=request_options,
        )

        # Log submission
        job_name = parameters["name"]
        job_id = response.job_id
        user_msg = response.job_submit_user_msg
        if user_msg:
            logger.info(
                "SLURM job '%s' submitted (Job ID: %s): %s", job_name, job_id, user_msg
            )
        else:
            logger.info("SLURM job '%s' submitted (Job ID: %s)", job_name, job_id)

        return job_id

    def _submit_ensure_name(self, parameters: dict) -> None:
        if not parameters.get("name"):
            parameters["name"] = defaults.JOB_NAME

    def _submit_ensure_std(self, parameters: dict) -> None:
        parameters["standard_input"] = "/dev/null"
        if self._log_directory:
            os_utils.makedirs(self._log_directory)
            job_name = parameters["name"]
            filetemplate = f"{self._log_directory}/{job_name}.%j"
            if self._std_split:
                parameters["standard_output"] = f"{filetemplate}.out"
                parameters["standard_error"] = f"{filetemplate}.err"
            else:
                parameters["standard_output"] = f"{filetemplate}.outerr"
        else:
            parameters["standard_output"] = "/dev/null"
            parameters["standard_error"] = "/dev/null"

    def _submit_ensure_environment(self, parameters: dict) -> None:
        if parameters.get("environment"):
            return
        parameters["environment"] = {"_DUMMY_VAR": "dummy_value"}

    def _submit_ensure_wd(self, parameters: dict) -> None:
        """The slurm user does not necessarily have a home directory
        and if it has, "~" does not work.
        """
        if not parameters.get("current_working_directory"):
            parameters["current_working_directory"] = defaults.SLURM_TEMPDIR

    def cancel_job(self, job_id: int, request_options: Optional[dict] = None) -> None:
        request_options = utils.merge_mappings(
            request_options, {"path_params": {"job_id": job_id}}
        )

        _ = self.delete(
            f"/slurm/{self._api_version_str}/job/{{job_id}}",
            request_options=request_options,
        )

    def clean_job_artifacts(
        self,
        job_id: int,
        job_properties: Optional[BaseModel] = None,
        request_options: Optional[dict] = None,
    ) -> None:
        if job_properties is None:
            job_properties = self.get_job_properties(
                job_id, raise_on_error=False, request_options=request_options
            )
        if job_properties is None:
            return
        slurm_logs.remove_slurm_log(job_properties, stderr=False)
        slurm_logs.remove_slurm_log(job_properties, stderr=True)

    @contextmanager
    def clean_job_artifacts_context(
        self,
        job_id: int,
        job_properties: Optional[BaseModel] = None,
        request_options: Optional[dict] = None,
    ) -> Generator[None, None, None]:
        try:
            yield
        finally:
            self.clean_job_artifacts(
                job_id,
                job_properties=job_properties,
                request_options=request_options,
            )

    def get_job_properties(
        self, job_id: int, raise_on_error=True, request_options: Optional[dict] = None
    ) -> Optional[BaseModel]:
        request_options = utils.merge_mappings(
            request_options, {"path_params": {"job_id": job_id}}
        )

        response = self.get(
            f"/slurm/{self._api_version_str}/job/{{job_id}}",
            request_options=request_options,
            raise_on_error=raise_on_error,
        )

        if response is None:
            return
        if not response.jobs:
            return
        return response.jobs[0]

    def get_all_job_properties(
        self,
        raise_on_error=True,
        request_options: Optional[dict] = None,
        filter: Optional[dict] = None,
        update_time: Optional[datetime.datetime] = None,
        all_users: bool = False,
    ) -> List[BaseModel]:
        # Server-side filter
        if update_time:
            request_options = utils.merge_mappings(
                request_options,
                {"query_params": {"update_time": update_time.timestamp()}},
            )

        response = self.get(
            f"/slurm/{self._api_version_str}/jobs",
            request_options=request_options,
            raise_on_error=raise_on_error,
        )
        if response is None or not response.jobs:
            return list()

        # Client-side filter
        if not filter:
            filter = dict()
        if not all_users:
            filter.setdefault("user_name", self._auth.user_name)

        return [
            job_properties
            for job_properties in response.jobs
            if self._is_part_of(job_properties, filter)
        ]

    def _is_part_of(self, obj: Any, subobj: Any) -> bool:
        """Check that `subobj` is a part of `obj`."""
        if isinstance(subobj, dict):
            # subobj must be a sub-tree of obj
            if not isinstance(obj, (dict, BaseModel)):
                return False
            for subobj_key, subobj_value in subobj.items():
                if isinstance(obj, BaseModel):
                    if not hasattr(obj, subobj_key):
                        return False
                    obj_value = getattr(obj, subobj_key)
                else:
                    if subobj_key not in obj:
                        return False
                    obj_value = obj[subobj_key]
                if not self._is_part_of(obj_value, subobj_value):
                    return False
            return True
        elif isinstance(subobj, list):
            # subobj must be equal to obj[:len(subobj)]
            if not isinstance(obj, list):
                return False
            if len(subobj) > len(obj):
                return False
            for obj_item, subobj_item in zip(obj, subobj):
                if not self._is_part_of(obj_item, subobj_item):
                    return False
            return True
        else:
            return obj == subobj

    def get_status(self, job_id: int, request_options: Optional[dict] = None) -> str:
        job_properties = self.get_job_properties(
            job_id, request_options=request_options
        )
        return slurm_response.slurm_job_state(job_properties)

    def get_full_status(
        self, job_id: int, request_options: Optional[dict] = None
    ) -> dict:
        job_properties = self.get_job_properties(
            job_id, request_options=request_options
        )
        return slurm_response.slurm_job_full_state(job_properties)

    def is_finished(self, job_id: int) -> bool:
        """Returns `True` when the job is completely finished."""
        return self.get_status(job_id) in slurm_response.FINISHED_STATES

    def is_finishing(self, job_id: int, request_options: Optional[dict] = None) -> bool:
        """Returns `True` when the job is finished by Slurm is still finalizing."""
        return self.get_status(job_id) in slurm_response.FINISHING_STATES

    def wait_finished(self, job_id: int, **kw) -> str:
        """Wait until a job is finished and return the state."""
        return self.wait_states(job_id, slurm_response.FINISHED_STATES, **kw)

    def wait_finishing(self, job_id: int, **kw) -> str:
        """Wait until a job is finished or in the process of finishing and return the state."""
        return self.wait_states(job_id, slurm_response.FINISHING_STATES, **kw)

    def wait_states(
        self,
        job_id: int,
        states: Tuple[str],
        progress: bool = False,
        timeout: Optional[float] = None,
        period: float = 0.5,
    ) -> str:
        """Wait until a job state is reached and return the state."""
        job_state = None
        t0 = time.time()
        while job_state not in states:
            job_state = self.get_status(job_id)
            time.sleep(period)
            if progress:
                print(".", end="", flush=True)
            if timeout is not None and (time.time() - t0) > timeout:
                raise TimeoutError
        status = self.get_full_status(job_id)
        logger.info("Job '%s' finished: %s", job_id, pformat(status))
        return job_state

    def print_stdout_stderr(
        self, job_id: int, request_options: Optional[dict] = None
    ) -> None:
        job_properties = self.get_job_properties(
            job_id, request_options=request_options, raise_on_error=False
        )
        if job_properties is None:
            return

        stdout, stderr = slurm_logs.get_stdout_stderr(job_properties)
        print()
        print("".join(stdout.lines()))
        if stderr is not None:
            print("".join(stderr.lines()))

    def log_stdout_stderr(
        self,
        job_id: int,
        logger: Optional[logging.Logger] = None,
        level: int = logging.INFO,
        request_options: Optional[dict] = None,
        **log_options,
    ) -> None:
        job_properties = self.get_job_properties(
            job_id, request_options=request_options, raise_on_error=False
        )
        if job_properties is None:
            return

        stdout, stderr = slurm_logs.get_stdout_stderr(job_properties)
        stdout = "".join(stdout.lines())

        if logger is None:
            logger = logger
        if stderr is None:
            logger.log(level, "\n%s", stdout, **log_options)
        else:
            stderr = "".join(stderr.lines())
            logger.log(level, "\n%s\n%s", stdout, stderr, **log_options)

    @contextmanager
    def redirect_stdout_stderr(
        self, job_id: int, request_options: Optional[dict] = None
    ) -> Generator[None, None, None]:
        """Redirect logs files to the local root logger within this context."""
        job_properties = self.get_job_properties(
            job_id, raise_on_error=False, request_options=request_options
        )
        if job_properties is None:
            yield
            return

        ofilename = slurm_logs.slurm_log_filename(
            job_properties, job_properties.standard_output
        )
        efilename = slurm_logs.slurm_log_filename(
            job_properties, job_properties.standard_error
        )
        with log_file_monitor_context([ofilename, efilename]):
            yield

    def get(self, path: str, **kwargs) -> Optional[BaseModel]:
        return self.request("GET", path, **kwargs)

    def delete(self, path: str, **kwargs) -> Optional[BaseModel]:
        return self.request("DELETE", path, **kwargs)

    def post(self, path: str, **kwargs) -> Optional[BaseModel]:
        return self.request("POST", path, **kwargs)

    def request(
        self,
        method: str,
        path: str,
        request_options: Optional[dict] = None,
        raise_on_error: bool = True,
    ) -> Optional[BaseModel]:
        request_options = self._merged_request_options(request_options)
        try:
            return slurm_access.validated_slurm_request(
                method, self._base_url, path, **request_options
            )
        except Exception as e:
            headers = self._auth.retry_headers()
            if not headers:
                if raise_on_error:
                    raise
                logger.debug(str(e))
                return None
            request_options["headers"] = headers

        try:
            return slurm_access.validated_slurm_request(
                method, self._base_url, path, **request_options
            )
        except Exception as e:
            if raise_on_error:
                raise
            logger.debug(str(e))
        return None

    def server_has_api(
        self, request_options: Optional[dict] = None
    ) -> Union[bool, str, List[str]]:
        request_options = self._merged_request_options(request_options)

        openapi_spec = None
        try:
            openapi_spec = slurm_access.slurm_request(
                "GET", self._base_url, "/openapi", **request_options
            )
        except Exception:
            headers = self._auth.retry_headers()
            if not headers:
                raise
            request_options["headers"] = headers
        if openapi_spec is None:
            openapi_spec = slurm_access.slurm_request(
                "GET", self._base_url, "/openapi", **request_options
            )

        versions = slurm_response.extract_slurm_api_versions(openapi_spec)
        has_api = self._api_version_tuple in versions
        versions = [slurm_access.create_version(v) for v in versions]
        return has_api, self._api_version_str, versions

    def _merged_request_options(
        self,
        request_options: Optional[dict] = None,
    ) -> Dict[str, Any]:
        # self._auth > request_options > self._request_options
        headers = {"headers": self._auth.headers()}
        request_options = utils.merge_mappings(request_options, headers)
        request_options = utils.merge_mappings(self._request_options, request_options)
        return request_options


def _parse_slurm_rest_url(slurm_url: str) -> str:
    """
    :raises SlurmInvalidUrlError:
    """
    parsed = urlparse(slurm_url)
    if not parsed.scheme:
        raise SlurmInvalidUrlError(
            f"Empty SLURM REST URL scheme: {parsed.scheme}. Expected 'http' or 'https'."
        )

    if not parsed.hostname:
        raise SlurmInvalidUrlError("Missing hostname.")

    if parsed.port is None:
        if parsed.scheme == "https":
            default_port = 443
        elif parsed.scheme == "http":
            default_port = 6820
        else:
            default_port = None
        if default_port:
            parsed = url_utils.set_url_port(parsed, default_port)

    return urlunparse(parsed)



import os
import sys
import json
import time
import uuid
import errno
import logging
import pathlib
import threading
import functools
import subprocess
from urllib.parse import urlparse
from typing import Optional, Dict, Tuple, Union, List
from unittest.mock import patch
from contextlib import contextmanager
from concurrent.futures import ThreadPoolExecutor

import pytest
import requests

logger = logging.getLogger(__name__)

_MAX_JOBS = 50


@contextmanager
def mock_slurm_clients(tmp_path: pathlib.Path, slurm_parameters: Dict[str, str]):
    # Ensure `pyslurmutils` is not imported when `pyslurmutils_pytest_plugin` is imported.
    # When a package is imported before the `pytest_cov` coverage starts a `CoverageWarning`
    # is raised at the end of testing `pyslurmutils` itself.
    from pyslurmutils.client import defaults
    from pyslurmutils.client.rest.api import slurm_access

    last_job_id = 0
    jobs = dict()
    api_version_str = slurm_parameters["api_version"] or defaults.DEFAULT_API_VERSION
    api_version_tuple = slurm_access.parse_version(api_version_str)

    class MockResponse(requests.Response):
        def __init__(self, json_data, status_code: int):
            super().__init__()
            self.encoding = "utf-8"
            self._content = json.dumps(json_data).encode(self.encoding)
            self.status_code = status_code
            self.headers["Content-Type"] = "application/json"

    def _mock_request(
        method: str, url: str, json: Optional[dict] = None, **_
    ) -> MockResponse:
        nonlocal last_job_id

        parsed_url = urlparse(url)
        assert parsed_url.scheme == "http", url
        assert parsed_url.hostname == "mockhost", url
        assert parsed_url.port == 6820, url
        assert parsed_url.username is None, url
        path = parsed_url.path

        if (method, path) == ("GET", "/openapi"):
            return MockResponse(
                {
                    "paths": {
                        f"/slurm/{api_version_str}/job/submit": {},
                    }
                },
                200,
            )

        if method == "GET" and path.startswith(f"/slurm/{api_version_str}/job/"):
            job_id = int(path.split("/")[-1])
            job = jobs.get(job_id)
            if job is None:
                return MockResponse(
                    {
                        "last_backfill": {"set": False, "infinite": False, "number": 0},
                        "last_update": {"set": False, "infinite": False, "number": 0},
                        "warnings": [],
                        "errors": [],
                    },
                    404,
                )
            job_exec_info = job["job_exec_info"]
            logger.debug(
                "SLURM mock backend: get job %s %s", job_id, job_exec_info["job_state"]
            )
            return MockResponse(
                {
                    "jobs": [job_exec_info],
                    "last_backfill": {"set": False, "infinite": False, "number": 0},
                    "last_update": {"set": False, "infinite": False, "number": 0},
                    "warnings": [],
                    "errors": [],
                },
                200,
            )

        if method == "GET" and path == f"/slurm/{api_version_str}/jobs":
            logger.debug("SLURM mock backend: get jobs (# %d)", len(jobs))
            return MockResponse(
                {
                    "jobs": [job["job_exec_info"] for job in jobs.values()],
                    "last_backfill": {"set": False, "infinite": False, "number": 0},
                    "last_update": {"set": False, "infinite": False, "number": 0},
                    "warnings": [],
                    "errors": [],
                },
                200,
            )

        if method == "GET":
            raise NotImplementedError(path)

        if method == "DELETE" and path.startswith(f"/slurm/{api_version_str}/job/"):
            job_id = int(path.split("/")[-1])
            job = jobs.get(job_id)
            if job is None:
                return MockResponse(
                    {
                        "warnings": [],
                        "errors": [],
                    },
                    404,
                )
            job_exec_info = job["job_exec_info"]
            with job["job_state_lock"]:
                job_exec_info["job_state"] = _job_state("CANCELLED", api_version_tuple)
                logger.debug("SLURM mock backend: cancel job %s", last_job_id)
                return MockResponse(
                    {
                        "warnings": [],
                        "errors": [],
                    },
                    200,
                )

        if method == "DELETE":
            raise NotImplementedError(path)

        if path == f"/slurm/{api_version_str}/job/submit":
            last_job_id += 1
            job_exec_info = {
                "job_id": last_job_id,
                "job_state": _job_state("PENDING", api_version_tuple),
                **json["job"],
                "user_name": slurm_parameters["user_name"],
                "warnings": [],
                "errors": [],
            }
            job = {"job_exec_info": job_exec_info, "job_state_lock": threading.Lock()}
            jobs[last_job_id] = job
            logger.debug("SLURM mock backend: recieved job %s", last_job_id)

            if api_version_tuple < (0, 0, 39):
                script = json["script"]
            else:
                script = json["job"]["script"]

            lines = script.split("\n")
            shebang = lines[0]
            if "bash" in shebang:
                # bash script
                if sys.platform == "win32":
                    pytest.skip("bash script does not run on windows")
            elif "#!powershell" in shebang:
                # powershell script
                if sys.platform != "win32":
                    pytest.skip(
                        "windows batch script does not run on non-windows machine"
                    )
            elif "python" in shebang:
                # executable python script
                pass
            else:
                assert False, f"Unknown script starting with '{shebang}'"

            future = slurm_queue.submit(
                _job_main, script, json, job, tmp_path, api_version_tuple
            )
            futures.append(future)
            return MockResponse(job_exec_info, 200)

        if method == "POST":
            raise NotImplementedError(path)

    futures = []
    with ThreadPoolExecutor(max_workers=_MAX_JOBS) as slurm_queue:
        with patch("requests.request", side_effect=_mock_request):
            yield
            for future in futures:
                _ = future.result()


def _absorb_exception(function):
    @functools.wraps(function)
    def inner(*args, **kwargs):
        try:
            return function(*args, **kwargs)
        except Exception as e:
            logger.exception(f"Job execution failed: {e}")

    return inner


@_absorb_exception
def _job_main(
    script: str,
    json: dict,
    job: dict,
    tmp_path: pathlib.Path,
    api_version: Tuple[int, int, int],
) -> None:
    job_state_lock = job["job_state_lock"]
    job_exec_info = job["job_exec_info"]
    cmd = []
    lines = script.split("\n")

    shebang = lines[0]
    if "bash" in shebang:
        # bash script
        suffix = ".sh"
    elif "#!powershell" in shebang:
        # powershell script
        suffix = ".ps1"
        cmd = ["powershell", "-NoProfile", "-ExecutionPolicy", "Bypass", "-File"]
    elif "python" in shebang:
        # executable python script
        suffix = ".py"
        if sys.platform == "win32":
            lines.pop(0)
            cmd = [sys.executable]
    else:
        assert False, f"Unknown script starting with '{shebang}'"

    filename = tmp_path / f"tmp_{uuid.uuid4().hex}{suffix}"
    with open(filename, "w") as script:
        script.write("\n".join(lines))
        script.flush()
        os.fsync(script.fileno())
        os.chmod(filename, 0o755)

    filename = str(filename)
    logger.info("Created slurm script '%s'", filename)
    cmd.append(filename)

    env = dict(os.environ)
    for keyvalue in json["job"].get("environment", list()):
        key, _, value = keyvalue.partition("=")
        env[key] = value
    env["SLURM_JOB_ID"] = str(job_exec_info["job_id"])
    env["_TEST_TMPDIR"] = str(tmp_path)

    standard_output = json["job"].get("standard_output")
    standard_error = json["job"].get("standard_error")
    same_standard_output = standard_error == standard_output

    if standard_output == "/dev/null":
        stdout = subprocess.DEVNULL
        standard_output = None
    elif standard_output is None:
        stdout = None
    else:
        stdout = subprocess.PIPE

    if standard_error == "/dev/null":
        stderr = subprocess.DEVNULL
        standard_error = None
    elif standard_error is None:
        stderr = None
    else:
        if same_standard_output:
            stderr = subprocess.STDOUT
            standard_error = None
        else:
            stderr = subprocess.PIPE

    with _Popen(
        cmd,
        stdout=stdout,
        stderr=stderr,
        stdin=subprocess.DEVNULL,
        env=env,
        cwd=os.getcwd(),
    ) as proc:
        job_id = job_exec_info["job_id"]

        cancelled_state = _job_state("CANCELLED", api_version)
        running_state = _job_state("RUNNING", api_version)

        def monitor_cancellation():
            while proc.poll() is None:
                with job_state_lock:
                    if job_exec_info["job_state"] == cancelled_state:
                        logger.debug(
                            "SLURM mock backend: cancel job %s ... (PID=%d)",
                            job_id,
                            proc.pid,
                        )
                        proc.terminate()
                        break
                    elif job_exec_info["job_state"] != running_state:
                        job_exec_info["job_state"] = running_state
                        logger.debug("SLURM mock backend: job %s started", job_id)
                time.sleep(1)

        # Start cancellation monitor thread
        cancel_thread = threading.Thread(target=monitor_cancellation, daemon=True)
        cancel_thread.start()
        try:
            try:
                outs, errs = proc.communicate(timeout=60)
            except subprocess.TimeoutExpired:
                proc.kill()
                outs, errs = proc.communicate()
                logger.warning("Job %s timed out and was killed", job_id)

            if outs:
                logger.debug(
                    "SLURM mock backend: job %s stdout\n%s", job_id, outs.decode()
                )
            if errs:
                logger.debug(
                    "SLURM mock backend: job %s stderr\n%s", job_id, errs.decode()
                )

            # Write output and error if not cancelled
            if job_exec_info["job_state"] == _job_state("CANCELLED", api_version):
                logger.debug("SLURM mock backend: job %s cancelled", job_id)
            else:
                if standard_output is not None:
                    outfile = standard_output.replace("%j", str(job_id))
                    with open(outfile, "wb") as f:
                        f.write(outs)

                if standard_error is not None:
                    errfile = standard_error.replace("%j", str(job_id))
                    with open(errfile, "wb") as f:
                        f.write(errs)

                if proc.returncode:
                    job_exec_info["job_state"] = _job_state("FAILED", api_version)
                    logger.debug("SLURM mock backend: job %s failed", job_id)
                else:
                    job_exec_info["job_state"] = _job_state("COMPLETED", api_version)
                    logger.debug("SLURM mock backend: job %s completed", job_id)
        finally:
            cancel_thread.join()


def _job_state(state: str, api_version: Tuple[int, int, int]) -> Union[str, List[str]]:
    if api_version < (0, 0, 40):
        return state
    return [state]


def _Popen(*args, **kwargs) -> subprocess.Popen:
    for attempt in range(3):
        try:
            proc = subprocess.Popen(*args, **kwargs)
            break
        except OSError as e:
            if e.errno == errno.ETXTBSY and attempt < 2:
                time.sleep(0.05 * (attempt + 1))
                continue
            raise
    return proc



import os
import sys
import abc
import pickle
import traceback
import logging
import uuid
import tempfile
import importlib
import shutil
import hashlib


def init_job():
    job_id = os.environ.get("SLURM_JOB_ID", "???")
    logging.basicConfig(
        level=logging.WARNING,
        stream=sys.stdout,
        format=f"[SLURM{job_id}] [%(levelname)s] [%(name)s] %(message)s",
    )


def main_task_loop(client):
    while not client.max_tasks_reached():
        result = None
        exc = None
        exc_info = None
        job_failure = None

        try:
            data = client.receive()
            if data is None:
                break
        except Exception as e:
            exc = e
            job_failure = "failed receiving result from the client"
            exc_info = _serialize_exception(exc)

        if not job_failure:
            result, exc, exc_info = _execute(*data)
            if not isinstance(exc, Exception) and isinstance(exc, BaseException):
                job_failure = f"base exception '{type(exc)}' causes the client to exit"

        try:
            client.send(result, exc_info)
        except Exception:
            logging.warning("failed sending result to the client", exc_info=True)
            break
        finally:
            if job_failure:
                logging.error(job_failure)
                raise exc
    _cleanup_cached_modules()


class Client:
    """Client-side of a connection. The server-side is `ewoksjob.client.job_io.local._connection_base.Connection`."""

    _HEADER_NBYTES = 64  # allows for 2**64 bytes (~18 EB) data transfer
    _MAX_CHUNK_NBYTES = 65536  # data transfer in chunks of 64 KB

    def __init__(self, in_name, out_name) -> None:
        self._task_counter = 0
        self._in_name = in_name
        self._out_name = out_name
        self._max_tasks, log_level = self._receive()
        logging.getLogger().setLevel(log_level)
        logging.debug("Python version: %s", sys.version)
        logging.debug("working directory: %s", os.getcwd())
        logging.debug("Process ID: %s", os.getpid())

    def __enter__(self) -> "Client":
        return self

    def __exit__(self, *_) -> None:
        self.close()

    def max_tasks_reached(self):
        return self._max_tasks is not None and self._task_counter >= self._max_tasks

    def receive(self):
        try:
            data = self._receive()
        finally:
            self._task_counter += 1
        return data

    def _receive(self):
        logging.debug("Waiting for data from %s ...", self._in_name)
        bheader = self._receive_nbytes(self._HEADER_NBYTES)
        nbytes = int.from_bytes(bheader, "big")
        logging.debug("Receiving %d bytes from %s ...", nbytes, self._in_name)
        bdata = self._receive_nbytes(nbytes)
        logging.debug("Data from %s received", self._in_name)
        return pickle.loads(bdata)

    def send(self, data, exc_info):
        bdata = pickle.dumps((data, exc_info))
        nbytes = len(bdata)
        logging.debug(
            "Sending data %s (%s bytes) to %s ...",
            type(data),
            nbytes,
            self._out_name,
        )
        bheader = nbytes.to_bytes(self._HEADER_NBYTES, "big")
        self._send_bytes(bheader)
        self._send_bytes(bdata)
        logging.debug("Data send to %s", self._out_name)

    @abc.abstractmethod
    def close(self):
        pass

    @abc.abstractmethod
    def _receive_nbytes(self, nbytes):
        pass

    @abc.abstractmethod
    def _send_bytes(self, bdata):
        pass


def _execute(task, args, kwargs):
    if isinstance(task, tuple):
        func_name, source_code = task
        logging.info("Executing task: %s", func_name)
        # logging.debug("Code:\n%s", source_code)
    else:
        func_name = task.__name__
        logging.info("Executing task: %s", func_name)

    logging.debug("Arguments: %s", args)
    logging.debug("Keyword arguments: %s", kwargs)

    if args is None:
        args = tuple()
    if kwargs is None:
        kwargs = dict()

    result = None
    exc = None
    exc_info = None

    try:
        if isinstance(task, tuple):
            module = _import_temp_module(source_code)
            try:
                func = getattr(module, func_name)
            except AttributeError:
                raise NameError(f"function '{func_name}' not found in client code")

            result = func(*args, **kwargs)
        else:
            result = task(*args, **kwargs)

        logging.info("Succeeded task: %s", func_name)
        exc = None
        exc_info = None

    except BaseException as e:
        logging.info("Failed task: %s", func_name)
        traceback.print_exc()
        result = None
        exc = e
        exc_info = _serialize_exception(exc)

    return result, exc, exc_info


_MODULES = {}


def _import_temp_module(source_code):
    """Create a temporary module with the source code."""
    key = hashlib.sha256(source_code.encode("utf-8")).hexdigest()

    if key in _MODULES:
        module, _ = _MODULES[key]
        return module

    temp_dir = tempfile.mkdtemp(
        prefix="pyslurmutils_", dir=os.environ.get("_TEST_TMPDIR")
    )

    module_name = f"pyslurmutils_client_code_{uuid.uuid4().hex}"
    module_path = os.path.join(temp_dir, f"{module_name}.py")
    with open(module_path, "w", encoding="utf-8") as f:
        f.write(source_code)

    logging.debug("Import temporary module: %s", module_path)
    sys.path.insert(0, temp_dir)
    module = importlib.import_module(module_name)

    _MODULES[key] = (module, temp_dir)

    return module


def _cleanup_cached_modules():
    for _, temp_dir in _MODULES.values():
        try:
            if os.path.isdir(temp_dir):
                shutil.rmtree(temp_dir)
        except Exception:
            logging.exception("Failed to remove temp module directory %s", temp_dir)
    _MODULES.clear()


def _serialize_exception(exc):
    chain = list()
    cause = None
    while exc:
        if exc.__cause__ is not None:
            next_cause = True
            chained_exc = exc.__cause__
        elif exc.__context__ is not None and not exc.__suppress_context__:
            next_cause = False
            chained_exc = exc.__context__
        else:
            next_cause = None
            chained_exc = None

        exc_class = type(exc)
        exc_class_string = f"{exc_class.__module__}.{exc_class.__name__}"
        exc_tb_string = "".join(
            traceback.format_exception(exc_class, exc, exc.__traceback__, chain=False)
        )

        chain.append((exc_class_string, str(exc), exc_tb_string, cause))
        exc = chained_exc
        cause = next_cause
    return chain



import time
import threading

from concurrent.futures import as_completed
from ..client.job_io.local._executor import RemoteExecutor


def test_remote_executor_nomax():
    # All tasks are executed sequentially (time = ntasks * sleep)
    with RemoteExecutor(conservative_scheduling=True) as executor:
        nworkers = _assert_executor(executor, ntasks=10)
        assert nworkers == 1

    # All tasks are executed in parallel (time = sleep)
    with RemoteExecutor() as executor:
        nworkers = _assert_executor(executor, ntasks=10)
        assert nworkers == 10

    # Sequential blocks of `max_tasks_per_worker tasks` (time = max_tasks_per_worker * sleep)
    with RemoteExecutor(
        max_tasks_per_worker=3, conservative_scheduling=True
    ) as executor:
        nworkers = _assert_executor(executor, ntasks=10)
        assert nworkers == 4

    # All tasks are executed in parallel (time = sleep)
    with RemoteExecutor(max_tasks_per_worker=1) as executor:
        nworkers = _assert_executor(executor, ntasks=10)
        assert nworkers == 10


def test_remote_executor_max_workers():
    # Tasks executed in parallel blocks of `max_workers` (time = ceil(ntasks/max_workers) * sleep)
    with RemoteExecutor(max_workers=3) as executor:
        nworkers = _assert_executor(executor, ntasks=10)
        assert nworkers == 3

    # Tasks executed in parallel blocks of `max_workers` (time = ceil(ntasks/max_workers) * sleep)
    with RemoteExecutor(max_workers=3, max_tasks_per_worker=2) as executor:
        nworkers = _assert_executor(executor, ntasks=10)
        assert nworkers > 3

    # Tasks executed in parallel blocks of `max_workers` (time = ceil(ntasks/max_workers) * sleep)
    with RemoteExecutor(
        max_workers=3, max_tasks_per_worker=2, lazy_scheduling=False
    ) as executor:
        nworkers = _assert_executor(executor, ntasks=10)
        assert nworkers > 3


def _assert_executor(executor, ntasks) -> int:
    futures_list = []
    data = list(range(ntasks))
    for i in data:
        future = executor.submit(_example_task, i)
        futures_list.append(future)

    results = list()
    thread_ids = set()
    for future in as_completed(futures_list):
        i, thread_id = future.result()
        results.append(i)
        thread_ids.add(thread_id)

    assert data == sorted(results)
    return len(thread_ids)


def _example_task(i):
    time.sleep(0.3)
    return i, id(threading.current_thread())



# This file is automatically generated.
# DO NOT MODIFY THIS FILE MANUALLY.

ENDPOINTS = {
    ("POST", "/slurm/v0.0.41/job/submit"): (
        "v0_0_41",
        {
            "path": "PostPathSlurmJobSubmit",
            "query": "PostQuerySlurmJobSubmit",
            "body": "PostBodySlurmJobSubmit",
            "200": "PostResponse200SlurmJobSubmit",
            "default": "PostResponseDefaultSlurmJobSubmit",
        },
    ),
    ("GET", "/slurm/v0.0.41/job/{job_id}"): (
        "v0_0_41",
        {
            "path": "GetPathSlurmJobJobid",
            "query": "GetQuerySlurmJobJobid",
            "body": "GetBodySlurmJobJobid",
            "200": "GetResponse200SlurmJobJobid",
            "default": "GetResponseDefaultSlurmJobJobid",
        },
    ),
    ("DELETE", "/slurm/v0.0.41/job/{job_id}"): (
        "v0_0_41",
        {
            "path": "DeletePathSlurmJobJobid",
            "query": "DeleteQuerySlurmJobJobid",
            "body": "DeleteBodySlurmJobJobid",
            "200": "DeleteResponse200SlurmJobJobid",
            "default": "DeleteResponseDefaultSlurmJobJobid",
        },
    ),
    ("GET", "/slurm/v0.0.41/jobs"): (
        "v0_0_41",
        {
            "path": "GetPathSlurmJobs",
            "query": "GetQuerySlurmJobs",
            "body": "GetBodySlurmJobs",
            "200": "GetResponse200SlurmJobs",
            "default": "GetResponseDefaultSlurmJobs",
        },
    ),
    ("POST", "/slurm/v0.0.40/job/submit"): (
        "v0_0_40",
        {
            "path": "PostPathSlurmJobSubmit",
            "query": "PostQuerySlurmJobSubmit",
            "body": "PostBodySlurmJobSubmit",
            "200": "PostResponse200SlurmJobSubmit",
            "default": "PostResponseDefaultSlurmJobSubmit",
        },
    ),
    ("GET", "/slurm/v0.0.40/job/{job_id}"): (
        "v0_0_40",
        {
            "path": "GetPathSlurmJobJobid",
            "query": "GetQuerySlurmJobJobid",
            "body": "GetBodySlurmJobJobid",
            "200": "GetResponse200SlurmJobJobid",
            "default": "GetResponseDefaultSlurmJobJobid",
        },
    ),
    ("DELETE", "/slurm/v0.0.40/job/{job_id}"): (
        "v0_0_40",
        {
            "path": "DeletePathSlurmJobJobid",
            "query": "DeleteQuerySlurmJobJobid",
            "body": "DeleteBodySlurmJobJobid",
            "200": "DeleteResponse200SlurmJobJobid",
            "default": "DeleteResponseDefaultSlurmJobJobid",
        },
    ),
    ("GET", "/slurm/v0.0.40/jobs"): (
        "v0_0_40",
        {
            "path": "GetPathSlurmJobs",
            "query": "GetQuerySlurmJobs",
            "body": "GetBodySlurmJobs",
            "200": "GetResponse200SlurmJobs",
            "default": "GetResponseDefaultSlurmJobs",
        },
    ),
    ("POST", "/slurm/v0.0.39/job/submit"): (
        "v0_0_39",
        {
            "path": "PostPathSlurmJobSubmit",
            "query": "PostQuerySlurmJobSubmit",
            "body": "PostBodySlurmJobSubmit",
            "200": "PostResponse200SlurmJobSubmit",
            "default": "PostResponseDefaultSlurmJobSubmit",
        },
    ),
    ("GET", "/slurm/v0.0.39/job/{job_id}"): (
        "v0_0_39",
        {
            "path": "GetPathSlurmJobJobid",
            "query": "GetQuerySlurmJobJobid",
            "body": "GetBodySlurmJobJobid",
            "200": "GetResponse200SlurmJobJobid",
            "default": "GetResponseDefaultSlurmJobJobid",
        },
    ),
    ("DELETE", "/slurm/v0.0.39/job/{job_id}"): (
        "v0_0_39",
        {
            "path": "DeletePathSlurmJobJobid",
            "query": "DeleteQuerySlurmJobJobid",
            "body": "DeleteBodySlurmJobJobid",
            "200": "DeleteResponse200SlurmJobJobid",
            "default": "DeleteResponseDefaultSlurmJobJobid",
        },
    ),
    ("GET", "/slurm/v0.0.39/jobs"): (
        "v0_0_39",
        {
            "path": "GetPathSlurmJobs",
            "query": "GetQuerySlurmJobs",
            "body": "GetBodySlurmJobs",
            "200": "GetResponse200SlurmJobs",
            "default": "GetResponseDefaultSlurmJobs",
        },
    ),
}



import pytest
from ..client import SlurmBaseRestClient
from ..client.errors import RemoteHttpError


@pytest.fixture
def slurm_base_client(slurm_client_kwargs) -> SlurmBaseRestClient:
    slurm_client_kwargs = {
        k: v for k, v in slurm_client_kwargs.items() if k != "pre_script"
    }
    return SlurmBaseRestClient(**slurm_client_kwargs)


def test_version(slurm_base_client):
    assert slurm_base_client.server_has_api()


def test_wrong_job_wait(slurm_base_client):
    job_id = 0
    with pytest.raises(RemoteHttpError):
        slurm_base_client.wait_finished(job_id)


def test_wrong_job_print(slurm_base_client):
    job_id = 0
    slurm_base_client.print_stdout_stderr(job_id)


def test_wrong_job_clean(slurm_base_client):
    job_id = 0
    slurm_base_client.clean_job_artifacts(job_id)



import pytest

from ..client import SlurmScriptRestClient


@pytest.fixture
def slurm_script_client(slurm_client_kwargs) -> SlurmScriptRestClient:
    slurm_client_kwargs = {
        k: v for k, v in slurm_client_kwargs.items() if k != "pre_script"
    }
    return SlurmScriptRestClient(**slurm_client_kwargs)


def test_script(slurm_script_client, slurm_parameters):
    nbefore = len(list(slurm_script_client.get_all_job_properties(update_time=None)))
    job_id = slurm_script_client.submit_script(_SUCCESS_SCRIPT)
    try:
        try:
            assert slurm_script_client.wait_finished(job_id, timeout=60) == "COMPLETED"
        finally:
            slurm_script_client.print_stdout_stderr(job_id)

        assert slurm_script_client.get_status(job_id) == "COMPLETED"
        nafter = len(list(slurm_script_client.get_all_job_properties(update_time=None)))
        if slurm_parameters["mock"]:
            # This might fail on a production SLURM queue
            assert (nafter - nbefore) == 1
    finally:
        slurm_script_client.clean_job_artifacts(job_id)


def test_failing_script(slurm_script_client):
    job_id = slurm_script_client.submit_script(_FAIL_SCRIPT)
    try:
        assert slurm_script_client.wait_finished(job_id, timeout=60) == "FAILED"
        slurm_script_client.print_stdout_stderr(job_id)
    finally:
        slurm_script_client.clean_job_artifacts(job_id)


def test_cancel_script(slurm_script_client):
    job_id = slurm_script_client.submit_script(_SUCCESS_SCRIPT)
    try:
        slurm_script_client.cancel_job(job_id)
        assert slurm_script_client.wait_finished(job_id, timeout=60) == "CANCELLED"
        slurm_script_client.print_stdout_stderr(job_id)
    finally:
        slurm_script_client.clean_job_artifacts(job_id)


def test_script_parameters(slurm_script_client):
    parameters = {"time_limit": "00:05:00", "environment": {"NAME1": "VAR1"}}
    metadata = {"NAME2": "VAR2"}
    job_id = slurm_script_client.submit_script(
        _SUCCESS_SCRIPT, parameters=parameters, metadata=metadata
    )
    try:
        try:
            assert slurm_script_client.wait_finished(job_id, timeout=60) == "COMPLETED"
        finally:
            slurm_script_client.print_stdout_stderr(job_id)

        assert slurm_script_client.get_status(job_id) == "COMPLETED"
        properties = slurm_script_client.get_job_properties(job_id)
    finally:
        slurm_script_client.clean_job_artifacts(job_id)

    assert properties.comment == '{"NAME2": "VAR2"}'
    assert properties.time_limit.model_dump() == {
        "infinite": False,
        "number": 5,
        "set": True,
    }


_SUCCESS_SCRIPT = """
echo 'Job started'
echo 'Job finished'
"""

_FAIL_SCRIPT = """
echo 'Job started'
exit 3
"""



import os
import time
import logging

from ._base import Client
from ._base import main_task_loop
from ._base import init_job


def main():
    init_job()
    with FileClient() as client:
        main_task_loop(client)


class FileClient(Client):
    def __init__(self) -> None:
        input_filename = os.environ["_PYSLURMUTILS_INFILE"]
        output_filename = os.environ["_PYSLURMUTILS_OUTFILE"]

        logging.debug("Connecting to '%s' ...", input_filename)
        input_dirname = os.path.dirname(input_filename)
        while True:
            try:
                _ = os.listdir(input_dirname)  # force NFS cache
                self._input_file = open(input_filename, "rb+")
                self._input_filename = input_filename
                break
            except FileNotFoundError:
                time.sleep(0.5)
        logging.debug("Connected to '%s'", input_filename)

        self._output_file = open(output_filename, "wb+")
        self._output_filename = output_filename

        super().__init__(input_filename, output_filename)

    def close(self) -> None:
        logging.debug("Closing connection ...")
        if self._input_file is not None:
            self._input_file.close()
            self._input_file = None
        if self._output_file is not None:
            self._output_file.close()
            self._output_file = None
        logging.debug("Connection closed")

    def _send_bytes(self, bdata: bytes) -> None:
        self._output_file.write(bdata)
        self._output_file.flush()

    def _receive_nbytes(self, nbytes: int) -> bytes:
        data = bytearray()
        remaining = nbytes
        block = min(nbytes, self._MAX_CHUNK_NBYTES)

        while remaining > 0:
            filesize = os.path.getsize(self._input_filename)
            pos = self._input_file.tell()
            available = filesize - pos

            if available <= 0:
                time.sleep(0.1)
            else:
                chunk = self._input_file.read(min(block, remaining))
                data.extend(chunk)
                remaining -= len(chunk)

        return bytes(data)



import os
import logging
from typing import Generator
from contextlib import contextmanager

logger = logging.getLogger(__name__)


def chmod(path: str, mode: int = 0o777) -> None:
    """
    Safely set permissions of `path`.
    """
    with _temporary_umask(0):
        try:
            os.chmod(path, mode)
        except Exception as exc:
            _log_path_warning(path, exc)


def makedirs(dirname: str, mode: int = 0o777) -> None:
    """
    Safely create directories recursively.
    """
    with _temporary_umask(0):
        try:
            os.makedirs(dirname, mode=mode, exist_ok=True)
        except Exception as exc:
            _log_path_warning(dirname, exc)


def nfs_cache_refresh(dirname) -> None:
    # _ = os.listdir(dirname)  # not enough
    # _ = subprocess.run(["ls", "-l", dirname], check=False, text=True, capture_output=True)
    try:
        with os.scandir(dirname) as it:
            for entry in it:
                _ = entry.stat()  # This forces a fresh stat call, similar to "ls -l"
    except Exception as exc:
        _log_path_warning(dirname, exc)


@contextmanager
def _temporary_umask(umask: int = 0) -> Generator[None, None, None]:
    """
    Context manager to temporarily set os.umask, restoring the original value afterward.
    """
    original_umask = None
    try:
        original_umask = os.umask(umask)
    except Exception as exc:
        logger.warning("Failed to set umask: %s", exc)
    try:
        yield
    finally:
        if original_umask is not None:
            try:
                os.umask(original_umask)
            except Exception as exc:
                logger.warning("Failed to restore umask: %s", exc)


def _log_path_warning(path: str, exc: Exception) -> None:
    """
    Log detailed information about a path when an operation on it fails.
    """
    try:
        path_exists = os.path.exists(path)
    except Exception as exists_e:
        logger.warning(
            "Failed operation on '%s': %s. Also failed to check if path exists: %s",
            path,
            exc,
            exists_e,
        )
        return

    if not path_exists:
        logger.warning("Failed operation on '%s': %s. Path does not exist.", path, exc)
        return

    try:
        st = os.stat(path)
        perms = oct(st.st_mode & 0o777)
        uid = st.st_uid
        gid = st.st_gid
        logger.warning(
            "Failed operation on '%s': %s. Path exists with permissions %s, UID %s, GID %s",
            path,
            exc,
            perms,
            uid,
            gid,
        )
    except Exception as stat_e:
        logger.warning(
            "Failed operation on '%s': %s. Also failed to stat existing path: %s",
            path,
            exc,
            stat_e,
        )



import json
from typing import Tuple, Any


def add_parameters(parser):
    parser.add_argument(
        "script",
        nargs="*",
        type=str,
        default=["echo This is a test"],
        help="Inline script",
    )
    parser.add_argument(
        "-a",
        "--all",
        action="store_true",
        help="All jobs when no job ID is provided",
    )
    parser.add_argument(
        "-sp",
        "--slurm-parameter",
        dest="parameters",
        action="append",
        default=[],
        metavar="NAME=VALUE",
        help="SLURM parameters",
    )
    parser.add_argument(
        "-w",
        "--wait",
        action="store_true",
        help="Wait until the job is finished",
    )


def apply_parameters(args):
    args.parameters = dict(parse_parameter(p) for p in args.parameters)


def parse_parameter(option: str) -> Tuple[str, Any]:
    option, _, value = option.partition("=")
    return option, parse_value(value)


def parse_value(value: str) -> Any:
    try:
        return json.loads(value)
    except Exception:
        return value



"""Monitor SLURM jobs and status"""

import sys
import time
import argparse
import datetime
import logging

from tabulate import tabulate
from pydantic import BaseModel

from .cli import common as common_cli
from .cli import status as status_cli
from .cli import cancel as cancel_cli
from .cli import submit as submit_cli
from .cli import diff as diff_cli
from ..client import defaults
from ..client import SlurmScriptRestClient
from ..client.rest.api import slurm_response


logger = logging.getLogger(__name__)


def create_argument_parser():
    parser = argparse.ArgumentParser(
        description="SLURM Job Monitor", prog="pyslurmutils"
    )
    subparsers = parser.add_subparsers(help="Commands", dest="command")

    check = subparsers.add_parser("check", help="Check Slurm connection")
    common_cli.add_parameters(check)

    subparser = subparsers.add_parser("status", help="Job status")
    common_cli.add_parameters(subparser)
    status_cli.add_parameters(subparser)

    subparser = subparsers.add_parser("diff", help="Difference between two jobs")
    common_cli.add_parameters(subparser)
    diff_cli.add_parameters(subparser)

    subparser = subparsers.add_parser("cancel", help="Cancel jobs")
    common_cli.add_parameters(subparser)
    cancel_cli.add_parameters(subparser)

    subparser = subparsers.add_parser("submit", help="Submit job")
    common_cli.add_parameters(subparser)
    submit_cli.add_parameters(subparser)

    subparser = subparsers.add_parser("version", help="Slurm API version")
    common_cli.add_parameters(subparser)

    return parser


def main(argv=None):
    if argv is None:
        argv = sys.argv

    parser = create_argument_parser()

    args = parser.parse_args(argv[1:])

    if args.command == "status":
        command_status(args)
    elif args.command == "diff":
        command_diff(args)
    elif args.command == "check":
        command_check(args)
    elif args.command == "cancel":
        command_cancel(args)
    elif args.command == "submit":
        command_submit(args)
    elif args.command == "version":
        command_version(args)
    else:
        parser.print_help()
    return 0


def command_status(args):
    common_cli.apply_parameters(args)
    status_cli.apply_parameters(args)

    client = SlurmScriptRestClient(
        url=args.url,
        user_name=args.user_name,
        token=args.token,
        api_version=args.api_version,
        renewal_url=args.renewal_url,
        log_directory=args.log_directory,
    )
    for _ in _monitor_loop(args.interval):
        _print_jobs(client, args.jobid, args.all)
        if args.jobid:
            client.print_stdout_stderr(args.jobid)


def command_diff(args):
    common_cli.apply_parameters(args)
    diff_cli.apply_parameters(args)

    client = SlurmScriptRestClient(
        url=args.url,
        user_name=args.user_name,
        token=args.token,
        api_version=args.api_version,
        renewal_url=args.renewal_url,
        log_directory=args.log_directory,
    )

    _diff_jobs(client, args.jobid1, args.jobid2)


def command_check(args):
    common_cli.apply_parameters(args)
    client = SlurmScriptRestClient(
        url=args.url,
        user_name=args.user_name,
        token=args.token,
        api_version=args.api_version,
        renewal_url=args.renewal_url,
        log_directory=args.log_directory,
    )
    has_api, selected_api, versions = client.server_has_api()
    print(f"Selected SLURM API {selected_api}")
    print(f"Server supports {', '.join(versions)}")
    if has_api:
        print("OK!")
    else:
        raise RuntimeError(f"SLURM API {selected_api} is not supported by the server!")


def command_cancel(args):
    common_cli.apply_parameters(args)
    cancel_cli.apply_parameters(args)
    client = SlurmScriptRestClient(
        url=args.url,
        user_name=args.user_name,
        token=args.token,
        api_version=args.api_version,
        renewal_url=args.renewal_url,
        log_directory=args.log_directory,
    )
    job_ids = args.job_ids
    if not job_ids:
        job_ids = [prop["job_id"] for prop in client.get_all_job_properties()]
    for job_id in job_ids:
        client.cancel_job(job_id)


def command_submit(args):
    common_cli.apply_parameters(args)
    submit_cli.apply_parameters(args)
    client = SlurmScriptRestClient(
        url=args.url,
        user_name=args.user_name,
        token=args.token,
        api_version=args.api_version,
        renewal_url=args.renewal_url,
        parameters=args.parameters,
        log_directory=args.log_directory,
    )
    job_id = client.submit_script(args.script)
    print(f"SLURM job {job_id} started")
    if not args.wait:
        return
    try:
        client.wait_finished(job_id)
        client.print_stdout_stderr(job_id)
    finally:
        client.clean_job_artifacts(job_id)


def command_version(args):
    common_cli.apply_parameters(args)
    client = SlurmScriptRestClient(
        url=args.url,
        user_name=args.user_name,
        token=args.token,
        api_version=args.api_version,
        renewal_url=args.renewal_url,
    )
    print(f"Default SLURM API {defaults.DEFAULT_API_VERSION}")
    print(f"Selected SLURM API {client._api_version_str}")


def _monitor_loop(interval):
    try:
        if not interval:
            yield
            return
        while True:
            yield
            time.sleep(interval)
    except KeyboardInterrupt:
        pass


def _print_jobs(client, jobid, all_users):
    if jobid:
        jobs = [client.get_job_properties(jobid)]
    else:
        jobs = client.get_all_job_properties(all_users=all_users)

    columns = {
        "ID": (_as_str, ("job_id",)),
        "Name": (_as_str, ("name", 30)),
        "State": (_job_state, ()),
        "User": (_as_str, ("user_name",)),
        "Limit": (_as_minutes, ("time_limit",)),
        "Submit": (_as_timestamp, ("submit_time",)),
        "Pendtime": (_time_diff, ("submit_time", "start_time", False)),
        "Runtime": (_time_diff, ("start_time", "end_time", True)),
        "Resources": (_resources, ("partition", "tres_req_str", "tres_alloc_str")),
    }

    rows = list()
    for job_properties in jobs:
        rows.append(
            [
                parser(job_properties, *parser_args)
                for parser, parser_args in columns.values()
            ]
        )
    if not rows:
        print("No jobs found!")
        return
    titles = list(columns)
    table = tabulate(rows, headers=titles)
    print(table)


def _as_str(job_properties: BaseModel, key: str, max_len: int = None) -> str:
    string = str(getattr(job_properties, key, "-"))
    if not max_len or len(string) <= max_len:
        return string
    return string[: max_len - 3] + "..."


def _job_state(job_properties: BaseModel) -> str:
    return slurm_response.slurm_job_state(job_properties)


def _resources(
    job_properties: BaseModel, partition: str, requested: str, allocated: str
) -> str:
    partition = getattr(job_properties, partition, None)
    requested = getattr(job_properties, requested, None)
    allocated = getattr(job_properties, allocated, None)
    return f"{partition}: {requested or allocated}"


def _as_timestamp(job_properties: BaseModel, epoch: str) -> datetime.datetime:
    epoch = getattr(job_properties, epoch, None)
    return slurm_response.slurm_unix_timestamp_or_now(epoch)


def _as_minutes(job_properties: BaseModel, minutes: str) -> datetime.timedelta:
    minutes = getattr(job_properties, minutes, None)
    return slurm_response.slurm_duration_minutes(minutes)


def _time_diff(
    job_properties: BaseModel, start_time: str, end_time: str, check_state: bool
) -> str:
    start_time = getattr(job_properties, start_time, None)
    end_time = getattr(job_properties, end_time, None)

    if check_state:
        state = slurm_response.slurm_job_state(job_properties)
        if state == "PENDING":
            return "-"
        if state not in slurm_response.FINISHING_STATES:
            end_time = None

    start_time = slurm_response.slurm_unix_timestamp_or_now(start_time)
    end_time = slurm_response.slurm_unix_timestamp_or_now(end_time)
    duration = end_time - start_time
    if duration.total_seconds() < 0:
        return "-"
    return str(duration)


def _diff_jobs(client, jobid1: int, jobid2: int):
    job1 = client.get_job_properties(jobid1)
    job2 = client.get_job_properties(jobid2)

    diffs = _recursive_diff(job1.model_dump(), job2.model_dump())

    if diffs:
        rows = [
            (path, _truncate_table_value(v1, 60), _truncate_table_value(v2, 60))
            for path, (v1, v2) in diffs.items()
        ]
        print(tabulate(rows, headers=["Path", "Job 1", "Job 2"], tablefmt="fancy_grid"))
    else:
        print("Jobs are identical!")


def _truncate_table_value(value, width: int) -> str:
    s = str(value)
    if len(s) <= width:
        return s
    return s[: width - 3] + "..."


def _recursive_diff(d1, d2, path=""):
    diffs = {}

    if isinstance(d1, dict) and isinstance(d2, dict):
        all_keys = set(d1) | set(d2)
        for key in all_keys:
            v1 = d1.get(key, "<MISSING>")
            v2 = d2.get(key, "<MISSING>")
            current_path = f"{path}.{key}" if path else key
            diffs.update(_recursive_diff(v1, v2, current_path))

    elif isinstance(d1, list) and isinstance(d2, list):
        try:
            d1 = sorted(d1)
            d2 = sorted(d2)
        except Exception:
            pass
        max_len = max(len(d1), len(d2))
        for i in range(max_len):
            v1 = d1[i] if i < len(d1) else "<MISSING>"
            v2 = d2[i] if i < len(d2) else "<MISSING>"
            current_path = f"{path}[{i}]"
            if isinstance(v1, (dict, list)) and isinstance(v2, (dict, list)):
                diffs.update(_recursive_diff(v1, v2, current_path))
            elif v1 != v2:
                diffs[current_path] = (v1, v2)

    elif d1 != d2:
        diffs[path] = (d1, d2)

    return diffs


if __name__ == "__main__":
    sys.exit(main())



import datetime
from typing import Any, Dict, List, Tuple, Union, Optional

_VersionType = Tuple[int, int, int]


def coerce_parameters(version: _VersionType, parameters: dict) -> None:
    for key in ("time_limit", "time_minimum"):
        if key in parameters:
            parameters[key] = _timedelta_in_minutes(version, parameters[key])

    if "environment" in parameters:
        parameters["environment"] = _environment(version, parameters["environment"])

    for key in ("memory_per_cpu", "memory_per_node"):
        if key in parameters:
            parameters[key] = _memory_in_megabytes(version, parameters[key])

    for key in ("memory_per_tres",):
        if key in parameters:
            parameters[key] = _tres_memory_in_megabytes(version, parameters[key])


def _integral_number(
    version: _VersionType,
    number: Any,
    lower_bound: Optional[int] = None,
    invalid: Optional[int] = None,
) -> Any:
    """Convert a number into Slurm API dict format.
    For example "10.4", "10", 10.4 and 10 all result in 10.
    """
    if isinstance(number, str):
        try:
            number = int(number)
        except Exception:
            try:
                number = float(number)
            except Exception:
                return number

    if isinstance(number, float):
        number = int(round(number))

    if isinstance(number, int):
        if lower_bound is not None:
            number = max(number, lower_bound)
        return {"number": number, "set": number != invalid, "infinite": False}

    return number


def _timedelta_in_minutes(version: _VersionType, delta_time: Any) -> Any:
    """Convert a Slurm time limit string "HH:MM:SS" or number into minutes."""
    if isinstance(delta_time, str):
        try:
            hours, minutes, seconds = map(int, delta_time.split(":"))
            delta_time = (
                datetime.timedelta(
                    hours=hours, minutes=minutes, seconds=seconds
                ).total_seconds()
                / 60.0
            )
        except Exception:
            return delta_time

    if isinstance(delta_time, (int, float)):
        return _integral_number(version, delta_time, lower_bound=1)

    return delta_time


def _environment(
    version: _VersionType, env: Dict[str, Any]
) -> Union[Dict[str, str], List[str]]:
    """
    Convert environment dict into proper Slurm API format.
    - For version < 0.0.39 → dict {k: str(v)}
    - For version >= 0.0.39 → list of "KEY=VALUE" strings
    """
    if version < (0, 0, 39):
        return {k: str(v) for k, v in env.items()}
    return [f"{k}={v}" for k, v in env.items()]


def _memory_in_megabytes(version: _VersionType, mem: Any) -> Any:
    """Convert memory specification into Slurm dict format in MB.

    Accepts:

      - int/float (already MB)
      - str with unit suffix (e.g., "1G", "512M")
    """
    if version < (0, 0, 39):
        return mem

    if isinstance(mem, str):
        mem = _parse_mem_string(mem)

    if isinstance(mem, (int, float)):
        return _integral_number(version, mem, lower_bound=1)

    return mem


def _tres_memory_in_megabytes(version: _VersionType, tres_mem: Any) -> Any:
    """Convert TRES memory string into a semicolon-delimited string with MB values.

    Example:
      "gres/gpu:2G;gres/mic:512M" → "gres/gpu:2048;gres/mic:512"
    """
    if version < (0, 0, 39):
        return tres_mem
    if not isinstance(tres_mem, str):
        return tres_mem

    tres_map: Dict[str, int] = {}
    for part in tres_mem.replace(" ", "").split(";"):
        if not part:
            continue
        if ":" not in part:
            return tres_mem
        tres, mem = part.split(":", 1)
        number = max(1, int(round(_parse_mem_string(mem))))
        tres_map[tres] = tres_map.get(tres, 0) + number

    return ";".join(f"{k}:{v}" for k, v in tres_map.items())


def _parse_mem_string(mem: str) -> float:
    """Convert a memory string like '2T', '2G', '512M', '100K' to MB (float)."""
    s = mem.strip().upper()
    if not s:
        return 0.0

    if s.endswith("TB"):
        return float(s[:-2]) * 1024 * 1024
    if s.endswith("T"):
        return float(s[:-1]) * 1024 * 1024
    if s.endswith("GB"):
        return float(s[:-2]) * 1024
    if s.endswith("G"):
        return float(s[:-1]) * 1024
    if s.endswith("MB"):
        return float(s[:-2])
    if s.endswith("M"):
        return float(s[:-1])
    if s.endswith("KB"):
        return float(s[:-2]) / 1024
    if s.endswith("K"):
        return float(s[:-1]) / 1024
    return float(s)



import pytest

from ..client.rest.api.slurm_params import _integral_number
from ..client.rest.api.slurm_params import coerce_parameters


def test_integral_number():
    assert _integral_number((0, 0, 41), 5) == {
        "number": 5,
        "set": True,
        "infinite": False,
    }
    assert _integral_number((0, 0, 41), 5.2) == {
        "number": 5,
        "set": True,
        "infinite": False,
    }
    assert _integral_number((0, 0, 41), 5.5) == {
        "number": 6,
        "set": True,
        "infinite": False,
    }
    assert _integral_number((0, 0, 41), "10") == {
        "number": 10,
        "set": True,
        "infinite": False,
    }
    assert _integral_number((0, 0, 41), "10.6") == {
        "number": 11,
        "set": True,
        "infinite": False,
    }
    assert _integral_number((0, 0, 41), "not-a-number") == "not-a-number"


def test_timedelta_in_minutes():
    parameters = {"time_limit": "01:00:00"}
    expected = {"time_limit": {"number": 60, "set": True, "infinite": False}}
    coerce_parameters((0, 0, 41), parameters)
    assert parameters == expected

    parameters = {"time_limit": 30.6}
    expected = {"time_limit": {"number": 31, "set": True, "infinite": False}}
    coerce_parameters((0, 0, 41), parameters)
    assert parameters == expected

    parameters = {"time_limit": "not-a-time"}
    expected = {"time_limit": "not-a-time"}
    coerce_parameters((0, 0, 41), parameters)
    assert parameters == expected


def test_environment_version():
    parameters = {"environment": {"A": 1, "B": "test"}}
    expected = {"environment": {"A": "1", "B": "test"}}
    coerce_parameters((0, 0, 38), parameters)
    assert parameters == expected

    parameters = {"environment": {"A": 1, "B": "test"}}
    expected = {"environment": ["A=1", "B=test"]}
    coerce_parameters((0, 0, 41), parameters)
    assert parameters == expected


@pytest.mark.parametrize(
    "mem, int_mb",
    [
        ("1G", 1024),
        ("1GB", 1024),
        ("0.5GB", 512),
        ("512M", 512),
        ("100K", 1),
        ("2048", 2048),
        (2048.2, 2048),
    ],
)
def test_memory_in_megabytes(mem, int_mb):
    parameters = {"memory_per_cpu": mem}
    expected = {"memory_per_cpu": {"number": int_mb, "set": True, "infinite": False}}
    coerce_parameters((0, 0, 41), parameters)
    assert parameters == expected


def test_tres_memory_in_megabytes():
    parameters = {"memory_per_tres": "gres/gpu:2G; gres/mic:512M"}
    expected = {"memory_per_tres": "gres/gpu:2048;gres/mic:512"}
    coerce_parameters((0, 0, 41), parameters)
    assert parameters == expected

    parameters = {"memory_per_tres": "gres/gpu:1G;gres/gpu:512M"}
    expected = {"memory_per_tres": "gres/gpu:1536"}
    coerce_parameters((0, 0, 41), parameters)
    assert parameters == expected

    parameters = {"memory_per_tres": "invalid_format"}
    expected = {"memory_per_tres": "invalid_format"}
    coerce_parameters((0, 0, 41), parameters)
    assert parameters == expected



"""SLURM API to submit, cancel and monitor scripts that start a python process
to establish a connection over which python functions can be executed."""

import os
import uuid
import logging
from typing import Optional, Union

from .script import SlurmScriptRestClient
from ..job_io.local import RemoteWorkerProxy
from .. import defaults

logger = logging.getLogger(__name__)


class SlurmPyConnRestClient(SlurmScriptRestClient):
    """SLURM API to submit, cancel and monitor scripts that start a python process
    to establish a connection over which python functions can be executed.
    This class does not contain any job-related state."""

    def __init__(
        self,
        url: str = "",
        user_name: str = "",
        token: str = "",
        api_version: str = "",
        renewal_url: str = "",
        parameters: Optional[dict] = None,
        log_directory: Optional[str] = None,
        std_split: Optional[bool] = False,
        request_options: Optional[dict] = None,
        pre_script: Optional[str] = None,
        post_script: Optional[str] = None,
        python_cmd: Optional[str] = None,
        use_os_environment: bool = True,
    ):
        """
        :param url: SLURM REST API URL (fallback to SLURM_URL env)
        :param user_name: SLURM username (fallback to SLURM_USER or system user)
        :param token: SLURM JWT token (fallback to SLURM_TOKEN env)
        :param api_version: SLURM API version (e.g. 'v0.0.41')
        :param renewal_url: Url for SLURM JWT token renewal (fallback to SLURM_RENEWAL_URL env)
        :param parameters: SLURM job parameters
        :param log_directory: SLURM log directory
        :param std_split: Split standard output and standard error
        :param request_options: GET, POST and DELETE options
        :param pre_script: Shell script to execute at the start of a job
        :param post_script: Shell script to execute at the end of a job
        :param python_cmd: Python command
        """
        self.pre_script = pre_script
        self.post_script = post_script
        self.python_cmd = python_cmd
        super().__init__(
            url=url,
            user_name=user_name,
            token=token,
            api_version=api_version,
            renewal_url=renewal_url,
            parameters=parameters,
            log_directory=log_directory,
            std_split=std_split,
            request_options=request_options,
            use_os_environment=use_os_environment,
        )

    def submit_script(
        self,
        worker_proxy: RemoteWorkerProxy,
        pre_script: Optional[str] = None,
        post_script: Optional[str] = None,
        python_cmd: Optional[str] = None,
        parameters: Optional[dict] = None,
        metadata: Optional[Union[str, dict]] = None,
        request_options: Optional[dict] = None,
    ) -> int:
        """Submit a script that will establish a connection initialized in the current process."""
        if parameters is None:
            parameters = dict()

        environment = parameters.setdefault("environment", dict())
        environment.update(worker_proxy.remote_environment)

        if not metadata:
            metadata = dict()
        metadata.update(worker_proxy.metadata)

        script = self._make_executable(
            worker_proxy.remote_script(),
            pre_script=pre_script,
            post_script=post_script,
            python_cmd=python_cmd,
        )

        return super().submit_script(
            script=script,
            parameters=parameters,
            metadata=metadata,
            request_options=request_options,
        )

    def _make_executable(
        self,
        python_script: str,
        pre_script: Optional[str] = None,
        post_script: Optional[str] = None,
        python_cmd: Optional[str] = None,
    ) -> str:
        """Create the code of a shell script that writes a Python script to disk,
        executes it, and then cleans it up.

        This is needed because scripts that use multiprocessing with the 'spawn'
        start method cannot reliably be executed via:
        - a shebang (e.g., `#!/usr/bin/env python3`)
        - stdin pipes (e.g., `python3 <<'EOF' ... EOF`)

        By writing the Python code to a real temporary file, we ensure that
        spawned child processes can import the main module safely.
        """
        pre_script = pre_script or self.pre_script or ""
        post_script = post_script or self.post_script or ""
        python_cmd = python_cmd or self.python_cmd or defaults.PYTHON_CMD

        # if not pre_script and not post_script:
        #     return f"#!/usr/bin/env {python_cmd}\n{python_script}"

        tmp_script = os.path.join(
            defaults.SLURM_TEMPDIR, f"pyslurmutils_main_{uuid.uuid4().hex}.py"
        )

        return _create_script(
            tmp_script, pre_script, python_cmd, python_script, post_script
        )


def _create_linux_script(
    tmp_script, pre_script, python_cmd, python_script, post_script
):
    """
    Create the code of a bash script that writes a Python script to disk,
    executes it, and then cleans it up.

    - Uses a `cat <<'PYTHONEOF' ... PYTHONEOF` block so that variables,
      backticks, and special characters in the Python code are not expanded
      or interpolated by the shell.
    - Use `exec` so the Python process inherits the same PID and will
      directly receive any signals (e.g., `SIGTERM`, `SIGINT`) sent to the
      Slurm job.
    """

    return f"""
{pre_script}
tmp_script="{tmp_script}"

cat > "$tmp_script" <<'PYTHONEOF'
{python_script}
PYTHONEOF

echo "Shell PID: $$"

exec {python_cmd} "$tmp_script"

rm -f "$tmp_script"

{post_script}
"""


def _create_windows_script(
    tmp_script, pre_script, python_cmd, python_script, post_script
):
    """
    Create the code of a Windows batch script that writes a Python script to disk,
    executes it, and then cleans it up.

    - Uses PowerShell heredoc (`@' ... '@`) to avoid problems with `echo`
      and special characters in the Python code.
    """

    return f"""
{pre_script}

$TmpScript = "{tmp_script}"

@'
{python_script}
'@ | Set-Content -Path $TmpScript -Encoding UTF8

Write-Output "PowerShell PID: $PID"

& {python_cmd} $TmpScript

Remove-Item -Force $TmpScript

{post_script}
"""


_create_script = _create_linux_script



from urllib.parse import ParseResult


def set_url_port(url: ParseResult, port: int) -> ParseResult:
    if url.username:
        netloc = f"{url.username}@{url.hostname}:{port}"
    else:
        netloc = f"{url.hostname}:{port}"

    return ParseResult(
        scheme=url.scheme,
        netloc=netloc,
        path=url.path,
        params=url.params,
        query=url.query,
        fragment=url.fragment,
    )



import logging
import weakref
import warnings
from pprint import pformat
from concurrent import futures
from contextlib import ExitStack
from contextlib import contextmanager
from typing import Optional, Any, Tuple, Type

from ..client import errors
from ..client import defaults
from ..client import SlurmPyConnRestClient
from ..client.rest.api import slurm_response
from ..client.job_io.local import RemoteExecutor
from ..client.job_io.local import ExecutorShutdown
from ..client.job_io.local import ExecuteContextReturnType
from ..client.job_io.local import RemoteWorkerProxy
from ..client.job_io.local import FileConnection
from ..client.job_io.local import TcpConnection


logger = logging.getLogger(__name__)


class SlurmRestFuture(futures.Future):
    def __init__(self) -> None:
        self._job_id = None
        self._slurm_client = None
        self._delayed_cancel_job = False
        super().__init__()

    def job_submitted(self, job_id: int, slurm_client: SlurmPyConnRestClient) -> None:
        """The SLURM job was submitted. Beware that the Slurm job may be running other tasks as well."""
        self._job_id = job_id
        self._slurm_client = weakref.proxy(slurm_client)
        if self._delayed_cancel_job:
            slurm_client.cancel_job(job_id)

    @property
    def job_id(self) -> Optional[int]:
        return self._job_id

    @property
    def slurm_client(self) -> Optional[SlurmPyConnRestClient]:
        return self._slurm_client

    def cancel_job(self) -> None:
        warnings.warn(
            "`cancel_job()` is deprecated and will be removed in a future release. Use `abort()` instead.",
            DeprecationWarning,
            stacklevel=2,
        )

    def abort(self) -> bool:
        """Cancel the Slurm job, even when it is already running. Beware that the Slurm job may be running other tasks as well."""
        slurm_client = self.slurm_client

        # The SLURM job was asked to be cancelled but didn't start yet: use `_delayed_cancel_job` to cancel it after it started.
        if slurm_client is None:
            self._delayed_cancel_job = True
        else:
            slurm_client.cancel_job(self.job_id)
        return self.aborted()

    def aborted(self) -> bool:
        slurm_client = self.slurm_client
        if slurm_client is None:
            return False
        status = slurm_client.get_status(self.job_id)
        return status == "CANCELLED"


class SlurmRestExecutor(RemoteExecutor):
    _FUTURE_CLASS = SlurmRestFuture

    def __init__(
        self,
        url: str = "",
        user_name: str = "",
        token: str = "",
        api_version: str = "",
        renewal_url: str = "",
        parameters: Optional[dict] = None,
        log_directory: Optional[str] = None,
        std_split: Optional[bool] = False,
        request_options: Optional[dict] = None,
        pre_script: Optional[str] = None,
        post_script: Optional[str] = None,
        python_cmd: Optional[str] = None,
        initializer: Optional[callable] = None,
        initargs: Optional[tuple] = None,
        initkwargs: Optional[tuple] = None,
        data_directory: Optional[str] = None,
        max_workers: Optional[int] = None,
        max_tasks_per_worker: Optional[int] = 1,
        lazy_scheduling: bool = True,
        conservative_scheduling: bool = False,
        cleanup_job_artifacts: bool = False,
        use_os_environment: bool = True,
    ):
        """
        :param url: SLURM REST API URL (fallback to SLURM_URL env)
        :param user_name: SLURM username (fallback to SLURM_USER or system user)
        :param token: SLURM JWT token (fallback to SLURM_TOKEN env)
        :param api_version: SLURM API version (e.g. 'v0.0.41')
        :param renewal_url: Url for SLURM JWT token renewal (fallback to SLURM_RENEWAL_URL env)
        :param parameters: SLURM job parameters
        :param log_directory: SLURM log directory
        :param std_split: Split standard output and standard error
        :param request_options: GET, POST and DELETE options
        :param pre_script: Shell script to execute at the start of a job
        :param post_script: Shell script to execute at the end of a job
        :param python_cmd: Python command
        :param initializer: execute when starting a job
        :param initargs: parameters for `initializer`
        :param initkwargs: parameters for `initializer`
        :param data_directory: communicate with the Slum job through files when specified
        :param max_workers: maximum number of Slum jobs that can run at any given time. `None` means unlimited.
        :param max_tasks_per_worker: maximum number of tasks each Slum job can receive before exiting. `None` means unlimited.
        :param lazy_scheduling: schedule SLURM jobs only when needed. Can only be disabled when `max_workers` is specified.
        :param conservative_scheduling: schedule the least amount of workers at the expense of tasks staying longer in the queue.
        :param cleanup_job_artifacts: cleanup job artifacts like logs.
        :param use_os_environment: Use ``SLURM_*`` environment variables
        """

        self._slurm_client = SlurmPyConnRestClient(
            url=url,
            user_name=user_name,
            token=token,
            api_version=api_version,
            renewal_url=renewal_url,
            log_directory=log_directory,
            parameters=parameters,
            std_split=std_split,
            request_options=request_options,
            pre_script=pre_script,
            post_script=post_script,
            python_cmd=python_cmd,
            use_os_environment=use_os_environment,
        )

        self._proxy_kwargs = {
            "max_tasks": max_tasks_per_worker,
            "initializer": initializer,
            "initargs": initargs,
            "initkwargs": initkwargs,
        }

        if data_directory:
            user_name = self._slurm_client._auth.user_name
            data_directory = str(data_directory).format(user_name=user_name)
            self._file_connection_kwargs = {
                "directory": data_directory,
                "basename": defaults.JOB_NAME,
            }
        else:
            self._file_connection_kwargs = None

        self._cleanup_job_artifacts = cleanup_job_artifacts

        super().__init__(
            max_workers=max_workers,
            max_tasks_per_worker=max_tasks_per_worker,
            lazy_scheduling=lazy_scheduling,
            conservative_scheduling=conservative_scheduling,
        )

    @contextmanager
    def execute_context(self) -> ExecuteContextReturnType:
        with ExitStack() as stack:

            job_id = None
            first_submit_kw = None

            def initialize(submit_kw):
                """
                Initialize SLURM worker: submit the SLURM job and initialize
                the communication with the worker.
                """
                nonlocal job_id, first_submit_kw
                first_submit_kw = submit_kw
                if submit_kw is None:
                    submit_kw = dict()
                job_id = self._slurm_client.submit_script(worker_proxy, **submit_kw)
                log_ctx = self._slurm_client.redirect_stdout_stderr(job_id)
                _ = stack.enter_context(log_ctx)

                if self._cleanup_job_artifacts:
                    cleanup_ctx = self._slurm_client.clean_job_artifacts_context(job_id)
                    _ = stack.enter_context(cleanup_ctx)

                worker_proxy.initialize()

            def execute(
                task: callable, args: tuple, kwargs: dict, future: SlurmRestFuture
            ) -> Any:
                """
                Send a task to the SLURM worker. Start the worker when not already running.
                """
                submit_kw = kwargs.pop(defaults.SLURM_ARGUMENTS_NAME, None)
                if job_id is None:
                    initialize(submit_kw)
                elif submit_kw != first_submit_kw:
                    logger.warning(
                        "SLURM submit arguments\n %s\n are ignored in favor of the arguments of the first task\n %s",
                        pformat(submit_kw),
                        pformat(first_submit_kw),
                    )

                future.job_submitted(job_id, self._slurm_client)

                try:
                    result, exc_info = worker_proxy.execute_without_reraise(
                        task, args=args, kwargs=kwargs
                    )
                except (ExecutorShutdown, errors.RemoteExit):
                    raise
                except Exception as ex:
                    exc_type, error_msg = status_error()
                    if exc_type:
                        raise exc_type(error_msg) from ex
                    raise

                if exc_info is not None:
                    errors.reraise_remote_exception_from_tb(exc_info)
                return result

            def status_error() -> Tuple[Optional[Type[Exception]], Optional[str]]:
                """Returns status exception class and message in case of a status error:
                remote Slurm job exited or local executor is shutting down.
                """
                if job_id is None:
                    return None, None
                try:
                    status = self._slurm_client.get_status(job_id)
                except Exception as e:
                    if self._shutdown_flag:
                        return ExecutorShutdown, "Slurm REST executor is shutting down"
                    logger.warning("failed getting the job state: %s", e, exc_info=True)
                    status = None
                if status in slurm_response.FINISHING_STATES:
                    return errors.RemoteExit, f"SLURM job {job_id} {status}"
                return None, None

            def raise_on_status_error() -> None:
                """Raise exception when there is a status error."""
                exc_type, error_message = status_error()
                if exc_type:
                    raise exc_type(error_message) from None

            def worker_exit_msg() -> Optional[str]:
                """Return exit message in case the executor worker must exit, i.e.
                when there is a status error."""
                _, exit_msg = status_error()
                return exit_msg

            if self._file_connection_kwargs:
                conn_ctx = FileConnection(
                    **self._file_connection_kwargs,
                    raise_on_status_error=raise_on_status_error,
                )
            else:
                conn_ctx = TcpConnection(raise_on_status_error=raise_on_status_error)
            connection = stack.enter_context(conn_ctx)
            proxy_ctx = RemoteWorkerProxy(connection, **self._proxy_kwargs)
            worker_proxy = stack.enter_context(proxy_ctx)

            if not self._lazy_scheduling:
                try:
                    initialize(None)
                except Exception as e:
                    logger.warning(
                        "SLURM worker initialization failed: %s", e, exc_info=True
                    )

            yield (execute, worker_exit_msg)



from typing import Any, Tuple, Optional

from ._proxy_base import WorkerProxy


class LocalWorkerProxy(WorkerProxy):
    """Worker proxy which executes tasks in the current process."""

    def __init__(
        self,
        initializer: Optional[callable] = None,
        initargs: Optional[tuple] = None,
        initkwargs: Optional[tuple] = None,
        max_tasks: Optional[int] = None,
    ) -> None:
        self._data = None
        super().__init__(
            initializer=initializer,
            initargs=initargs,
            initkwargs=initkwargs,
            max_tasks=max_tasks,
        )

    def _send(self, data: Any) -> None:
        self._data = data

    def _get_result(self) -> Tuple[Any, None]:
        try:
            if isinstance(self._data, tuple):
                task, args, kwargs = self._data
                if args is None:
                    args = tuple()
                if kwargs is None:
                    kwargs = dict()
                if isinstance(task, tuple):
                    func_name, source_code = task
                    lcls = locals()
                    dunder_name = lcls.get("__name__", None)
                    lcls["__name__"] = "__notmain__"
                    try:
                        exec(source_code, globals(), lcls)
                    finally:
                        if dunder_name:
                            lcls["__name__"] = dunder_name
                    func = lcls[func_name]
                    result = func(*args, **kwargs)
                else:
                    result = task(*args, **kwargs)
                return result, None
        finally:
            self._data = None
        return None, None



import os
import sys
import time

import pytest

from ..client import os_utils
from ..client.errors import RemoteExit
from ..concurrent.futures import SlurmRestExecutor


@pytest.fixture(params=["tcp", "file"])
def slurm_executor_kwargs(
    request, slurm_data_directory, slurm_client_kwargs, slurm_parameters
):
    if request.param == "tcp":
        slurm_data_directory = None
    elif not slurm_parameters["mock"]:
        request.node.add_marker(
            pytest.mark.xfail(reason="Distributed filesystems I/O is not reliable")
        )
    return {"data_directory": slurm_data_directory, **slurm_client_kwargs}


def test_rest_executor_submit(slurm_executor_kwargs):
    with SlurmRestExecutor(**slurm_executor_kwargs, max_workers=8) as slurm_executor:
        future1 = slurm_executor.submit(time.sleep, 0)
        future2 = slurm_executor.submit(sum, [1, 1])
        future3 = slurm_executor.submit(sum, [1, "a"])
        assert future1.result(timeout=60) is None
        assert future2.result(timeout=60) == 2
        with pytest.raises(TypeError):
            future3.result(timeout=60)


def test_rest_executor_map(slurm_executor_kwargs):
    with SlurmRestExecutor(**slurm_executor_kwargs, max_workers=8) as slurm_executor:
        results = [
            result
            for result in slurm_executor.map(sum, [[1, 1], [2, 2], [3, 3]], timeout=60)
        ]
    assert results == [2, 4, 6], str(results)


def test_rest_executor_initializer(slurm_executor_kwargs):
    with SlurmRestExecutor(
        **slurm_executor_kwargs,
        initializer=_initializer,
        max_workers=1,
        max_tasks_per_worker=5,
    ) as slurm_executor:
        ftls = [slurm_executor.submit(_increment_global_value) for _ in range(5)]
        results = {future.result(timeout=60) for future in ftls}
    assert results == set(range(1, 6)), str(results)


def _initializer():
    global GLOBAL_VALUE
    GLOBAL_VALUE = 0


def _increment_global_value():
    global GLOBAL_VALUE
    GLOBAL_VALUE += 1
    return GLOBAL_VALUE


def test_rest_executor_max_tasks_per_worker(slurm_executor_kwargs):
    with SlurmRestExecutor(
        **slurm_executor_kwargs, max_workers=8, max_tasks_per_worker=2
    ) as slurm_executor:
        # Note: the sleep time is needed to ensure jobs don't finish
        # before the submit for-loop is finished. In production this
        # does not matter but here we want to test `max_tasks_per_worker`.

        # Each worker executes one job.
        ftls = [slurm_executor.submit(_job_ident) for _ in range(8)]
        job_idents = {future.result(timeout=60) for future in ftls}
        assert len(job_idents) == 8, str(len(job_idents))

        # Each worker executes another job.
        ftls = [slurm_executor.submit(_job_ident) for _ in range(8)]
        job_idents |= {future.result(timeout=60) for future in ftls}
        assert len(job_idents) == 8, str(len(job_idents))

        # Each worker needs to be restarted because it has reached
        # the `max_tasks_per_worker` limit.
        ftls = [slurm_executor.submit(_job_ident) for _ in range(8)]
        job_idents |= {future.result(timeout=60) for future in ftls}
        assert len(job_idents) == 16, str(len(job_idents))


def _job_ident():
    time.sleep(0.3)
    return os.environ["SLURM_JOB_ID"]


@pytest.mark.parametrize("abort_delay", [0, 0.1, 1])
@pytest.mark.skipif(
    sys.platform == "win32", reason="Signal propagation not reliable on Windows"
)
def test_rest_executor_cancel(slurm_executor_kwargs, abort_delay):
    with SlurmRestExecutor(**slurm_executor_kwargs, max_workers=8) as slurm_executor:
        future = slurm_executor.submit(time.sleep, 10)
        if abort_delay:
            time.sleep(abort_delay)
        future.abort()

        with pytest.raises(RemoteExit, match=r"SLURM job \d+ CANCELLED"):
            future.result(timeout=60)


def test_slurm_tmp_path(slurm_executor_kwargs, slurm_tmp_path):
    print(slurm_tmp_path)
    filename = slurm_tmp_path / "test.txt"
    with SlurmRestExecutor(**slurm_executor_kwargs) as slurm_executor:
        future = slurm_executor.submit(_touch, str(filename))
        _ = future.result(timeout=60)
    os_utils.nfs_cache_refresh(slurm_tmp_path)
    assert filename.exists()


def _touch(filename):
    with open(filename, "w"):
        pass



import re
import datetime
from typing import Optional, Any, Tuple, List, Dict

from pydantic import BaseModel

_LOCAL_OFFSET = datetime.datetime.now().astimezone().utcoffset()

FINISHED_STATES = (
    "BOOT_FAIL",
    "CANCELLED",
    "COMPLETED",
    "DEADLINE",
    "FAILED",
    "NODE_FAIL",
    "OUT_OF_MEMORY",
    "PREEMPTED",
    "TIMEOUT",
)
FINISHING_STATES = "COMPLETING", *FINISHED_STATES

# https://slurm.schedmd.com/job_state_codes.html
#
# BOOT_FAIL     terminated due to node boot failure
# CANCELLED     cancelled by user or administrator
# COMPLETED     completed execution successfully; finished with an exit code of zero on all nodes
# DEADLINE      terminated due to reaching the latest acceptable start time specified for the job
# FAILED        completed execution unsuccessfully; non-zero exit code or other failure condition
# NODE_FAIL     terminated due to node failure
# OUT_OF_MEMORY experienced out of memory error
# PENDING       queued and waiting for initiation; will typically have a reason code specifying why it has not yet started
# PREEMPTED     terminated due to preemption; may transition to another state based on the configured PreemptMode and job characteristics
# RUNNING       allocated resources and executing
# SUSPENDED     allocated resources but execution suspended, such as from preemption or a direct request from an authorized user
# TIMEOUT       terminated due to reaching the time limit, such as those configured in slurm.conf or specified for the individual job


def slurm_integer(number: Any) -> Optional[int]:
    if isinstance(number, BaseModel):
        if number.set and not number.infinite and number.number is not None:
            return int(number.number)
    else:
        if number is not None:
            return int(number)


def slurm_float(number: Any) -> float:
    if isinstance(number, BaseModel):
        if number.infinite:
            return float("inf")
        if number.set and number.number is not None:
            return float(number.number)
        return float("nan")
    else:
        if number is not None:
            return float(number)


def slurm_unix_timestamp(number: Any) -> Optional[datetime.datetime]:
    epoch = slurm_integer(number)
    if epoch is not None:
        utc = datetime.datetime.utcfromtimestamp(epoch)
        local = utc + _LOCAL_OFFSET
        return local.astimezone()
    now = datetime.datetime.now().astimezone()
    return now.replace(microsecond=0)


def slurm_unix_timestamp_or_now(number: Any) -> datetime.datetime:
    dt = slurm_unix_timestamp(number)
    if dt is not None:
        return dt
    now = datetime.datetime.now().astimezone()
    return now.replace(microsecond=0)


def slurm_duration_minutes(number: Any) -> Optional[datetime.timedelta]:
    minutes = slurm_integer(number)
    if minutes is not None:
        return datetime.timedelta(minutes=minutes)


def slurm_error_messages(model: BaseModel) -> Tuple[List[str], str]:
    messages = []
    suffix = ""

    if isinstance(model.errors, list):
        errors = model.errors
    elif model.errors is not None:
        errors = [model.errors]
    else:
        errors = []

    for error in errors:
        if error.error_number in (5005, -1):
            # Errors observer when the SLURM token expired:
            #
            # [5005:Zero Bytes were transmitted or received]
            #  POST job/submit
            #  DELETE job/{job_id}
            #
            # [-1: Unspecified error]
            # GET job/{job_id}
            # GET jobs
            suffix = " (SLURM token expired?)"
        msg = (
            f"[{error.error_number}: {error.error}] {error.source}: {error.description}"
        )
        messages.append(msg)

    return messages, suffix


def slurm_warning_messages(method: str, path: str, model: BaseModel) -> List[str]:
    messages = list()
    for warning in model.warnings:
        msg = f"{method} {path} {warning.source}: {warning.description}"
        messages.append(msg)
    return messages


def slurm_job_state(job_properties: Optional[BaseModel]) -> str:
    return slurm_job_states(job_properties)[0]


def slurm_job_full_state(job_properties: Optional[BaseModel]) -> str:
    status = slurm_job_state(job_properties)
    if status == "NOJOB":
        return {
            "status": status,
            "description": "None",
            "reason": "None",
            "exit_code": "NaN",
        }
    return {
        "status": status,
        "description": job_properties.state_description,
        "reason": job_properties.state_reason,
        "exit_code": job_properties.exit_code,
    }


def slurm_job_states(job_properties: Optional[BaseModel]) -> List[str]:
    if job_properties is None:
        return ["NOJOB"]
    if isinstance(job_properties.job_state, list):
        return job_properties.job_state
    if isinstance(job_properties.job_state, str):
        return [job_properties.job_state]
    return ["UNKNOWN"]


def extract_slurm_api_versions(
    openapi_spec: Dict[str, Any],
) -> List[Tuple[int, int, int]]:
    """
    Get the list of supported SLURM REST API versions (newest first).
    """
    pattern = re.compile(r"\/slurm\/v([0-9\.]+)\/?.*")
    versions = set()
    for path in openapi_spec["paths"]:
        match = pattern.match(path)
        if match:
            version = match.group(1).split(".")
            version = tuple(map(int, version))
            versions.add(version)
    return sorted(versions)[::-1]



"""Remote execution: client-side API"""

from ._connection_base import Connection  # noqa F401
from ._connection_tcp import TcpConnection  # noqa F401
from ._connection_file import FileConnection  # noqa F401

from ._proxy_base import WorkerProxy  # noqa F401
from ._proxy_remote import RemoteWorkerProxy  # noqa F401
from ._proxy_local import LocalWorkerProxy  # noqa F401

from ._executor import RemoteExecutor  # noqa F401
from ._executor import ExecutorShutdown  # noqa F401
from ._executor import ExecuteContextReturnType  # noqa F401



# This file is automatically generated.
# DO NOT MODIFY THIS FILE MANUALLY.

from __future__ import annotations

from typing import Any, Dict, List, Optional

from pydantic import BaseModel, Field


class CronEntryLine(BaseModel):
    start: Optional[int] = Field(None, title="Start")
    end: Optional[int] = Field(None, title="End")


class DeleteBodySlurmJobJobid(BaseModel):
    pass


class DeletePathSlurmJobJobid(BaseModel):
    job_id: str = Field(..., title="Job Id")


class DeleteQuerySlurmJobJobid(BaseModel):
    signal: Optional[str] = Field(None, title="POSIX signal name")


class Error(BaseModel):
    error_number: Optional[int] = Field(None, title="Slurm internal error number")
    error: Optional[str] = Field(None, title="Error message")
    source: Optional[str] = Field(None, title="Where error occurred in the source")
    description: Optional[str] = Field(None, title="Explanation of cause of error")


class Float64NoVal(BaseModel):
    set: Optional[bool] = Field(
        None, title="True if number has been set. False if number is unset"
    )
    infinite: Optional[bool] = Field(
        None,
        title='True if number has been set to infinite. "set" and "number" will be ignored.',
    )
    number: Optional[float] = Field(
        None,
        title="If set is True the number will be set with value. Otherwise ignore number contents.",
    )


class GetBodySlurmJobJobid(BaseModel):
    pass


class GetBodySlurmJobs(BaseModel):
    pass


class GetPathSlurmJobJobid(BaseModel):
    job_id: str = Field(..., title="Job Id")


class GetPathSlurmJobs(BaseModel):
    pass


class GetQuerySlurmJobJobid(BaseModel):
    pass


class GetQuerySlurmJobs(BaseModel):
    update_time: Optional[int] = Field(None, title="Update Time")


class JobInfoPower(BaseModel):
    flags: Optional[List[Dict[str, Any]]] = Field(None, title="Flags")


class JobRes(BaseModel):
    nodes: Optional[str] = Field(None, title="Nodes")
    allocated_cores: Optional[int] = Field(None, title="Allocated Cores")
    allocated_cpus: Optional[int] = Field(None, title="Allocated Cpus")
    allocated_hosts: Optional[int] = Field(None, title="Allocated Hosts")
    allocated_nodes: Optional[List[Dict[str, Any]]] = Field(
        None, title="Allocated Nodes"
    )


class MetaPlugin(BaseModel):
    type: Optional[str] = Field(None, title="Type")
    name: Optional[str] = Field(None, title="Name")


class MetaSlurmVersion(BaseModel):
    major: Optional[int] = Field(None, title="Major")
    micro: Optional[int] = Field(None, title="Micro")
    minor: Optional[int] = Field(None, title="Minor")


class PostPathSlurmJobSubmit(BaseModel):
    pass


class PostQuerySlurmJobSubmit(BaseModel):
    pass


class Uint16NoVal(BaseModel):
    set: Optional[bool] = Field(
        None, title="True if number has been set. False if number is unset"
    )
    infinite: Optional[bool] = Field(
        None,
        title='True if number has been set to infinite. "set" and "number" will be ignored.',
    )
    number: Optional[int] = Field(
        None,
        title="If set is True the number will be set with value. Otherwise ignore number contents.",
    )


class Uint32NoVal(BaseModel):
    set: Optional[bool] = Field(
        None, title="True if number has been set. False if number is unset"
    )
    infinite: Optional[bool] = Field(
        None,
        title='True if number has been set to infinite. "set" and "number" will be ignored.',
    )
    number: Optional[int] = Field(
        None,
        title="If set is True the number will be set with value. Otherwise ignore number contents.",
    )


class Uint64NoVal(BaseModel):
    set: Optional[bool] = Field(
        None, title="True if number has been set. False if number is unset"
    )
    infinite: Optional[bool] = Field(
        None,
        title='True if number has been set to infinite. "set" and "number" will be ignored.',
    )
    number: Optional[int] = Field(
        None,
        title="If set is True the number will be set with value. Otherwise ignore number contents.",
    )


class Warning(BaseModel):
    warning: Optional[str] = Field(None, title="Earning message")
    source: Optional[str] = Field(None, title="Where error occurred in the source")
    description: Optional[str] = Field(None, title="Explanation of cause of error")


class CronEntry(BaseModel):
    flags: Optional[List[str]] = Field(None, title="Flags")
    minute: Optional[str] = Field(None, title="Minute")
    hour: Optional[str] = Field(None, title="Hour")
    day_of_month: Optional[str] = Field(None, title="Day Of Month")
    month: Optional[str] = Field(None, title="Month")
    day_of_week: Optional[str] = Field(None, title="Day Of Week")
    specification: Optional[str] = Field(None, title="Specification")
    command: Optional[str] = Field(None, title="Command")
    line: Optional[CronEntryLine] = Field(None, title="Cron entry line")


class JobDescMsg(BaseModel):
    account: Optional[str] = Field(None, title="Account")
    account_gather_frequency: Optional[str] = Field(
        None, title="Account Gather Frequency"
    )
    admin_comment: Optional[str] = Field(None, title="Admin Comment")
    allocation_node_list: Optional[str] = Field(None, title="Allocation Node List")
    allocation_node_port: Optional[int] = Field(None, title="Allocation Node Port")
    argv: Optional[List[str]] = Field(None, title="Argv")
    array: Optional[str] = Field(None, title="Array")
    batch_features: Optional[str] = Field(None, title="Batch Features")
    begin_time: Optional[int] = Field(None, title="Begin Time")
    flags: Optional[List[str]] = Field(None, title="Flags")
    burst_buffer: Optional[str] = Field(None, title="Burst Buffer")
    clusters: Optional[str] = Field(None, title="Clusters")
    cluster_constraint: Optional[str] = Field(None, title="Cluster Constraint")
    comment: Optional[str] = Field(None, title="Comment")
    contiguous: Optional[bool] = Field(None, title="Contiguous")
    container: Optional[str] = Field(None, title="Container")
    container_id: Optional[str] = Field(None, title="Container Id")
    core_specification: Optional[int] = Field(None, title="Core Specification")
    thread_specification: Optional[int] = Field(None, title="Thread Specification")
    cpu_binding: Optional[str] = Field(None, title="Cpu Binding")
    cpu_binding_flags: Optional[List[str]] = Field(None, title="Cpu Binding Flags")
    cpu_frequency: Optional[str] = Field(None, title="Cpu Frequency")
    cpus_per_tres: Optional[str] = Field(None, title="Cpus Per Tres")
    crontab: Optional[CronEntry] = Field(None, title="Cron entry")
    deadline: Optional[int] = Field(None, title="Deadline")
    delay_boot: Optional[int] = Field(None, title="Delay Boot")
    dependency: Optional[str] = Field(None, title="Dependency")
    end_time: Optional[int] = Field(None, title="End Time")
    environment: List[str] = Field(..., title="Environment")
    excluded_nodes: Optional[List[str]] = Field(None, title="Excluded Nodes")
    extra: Optional[str] = Field(None, title="Extra")
    constraints: Optional[str] = Field(None, title="Constraints")
    group_id: Optional[str] = Field(None, title="Group Id")
    hetjob_group: Optional[int] = Field(None, title="Hetjob Group")
    immediate: Optional[bool] = Field(None, title="Immediate")
    job_id: Optional[int] = Field(None, title="Job Id")
    kill_on_node_fail: Optional[bool] = Field(None, title="Kill On Node Fail")
    licenses: Optional[str] = Field(None, title="Licenses")
    mail_type: Optional[List[str]] = Field(None, title="Mail Type")
    mail_user: Optional[str] = Field(None, title="Mail User")
    mcs_label: Optional[str] = Field(None, title="Mcs Label")
    memory_binding: Optional[str] = Field(None, title="Memory Binding")
    memory_binding_type: Optional[List[str]] = Field(None, title="Memory Binding Type")
    memory_per_tres: Optional[str] = Field(None, title="Memory Per Tres")
    name: Optional[str] = Field(None, title="Name")
    network: Optional[str] = Field(None, title="Network")
    nice: Optional[int] = Field(None, title="Nice")
    tasks: Optional[int] = Field(None, title="Tasks")
    open_mode: Optional[List[str]] = Field(None, title="Open Mode")
    reserve_ports: Optional[int] = Field(None, title="Reserve Ports")
    overcommit: Optional[bool] = Field(None, title="Overcommit")
    partition: Optional[str] = Field(None, title="Partition")
    distribution_plane_size: Optional[int] = Field(
        None, title="Distribution Plane Size"
    )
    power_flags: Optional[List[Dict[str, Any]]] = Field(None, title="Power Flags")
    prefer: Optional[str] = Field(None, title="Prefer")
    hold: Optional[bool] = Field(None, title="Hold (true) or release (false) job")
    priority: Optional[Uint32NoVal] = Field(None, title="Request specific job priority")
    profile: Optional[List[str]] = Field(None, title="Profile")
    qos: Optional[str] = Field(None, title="Qos")
    reboot: Optional[bool] = Field(None, title="Reboot")
    required_nodes: Optional[List[str]] = Field(None, title="Required Nodes")
    requeue: Optional[bool] = Field(None, title="Requeue")
    reservation: Optional[str] = Field(None, title="Reservation")
    script: Optional[str] = Field(None, title="Script")
    shared: Optional[List[str]] = Field(None, title="Shared")
    exclusive: Optional[List[str]] = Field(None, title="Exclusive")
    oversubscribe: Optional[bool] = Field(None, title="Oversubscribe")
    site_factor: Optional[int] = Field(None, title="Site Factor")
    spank_environment: Optional[List[str]] = Field(None, title="Spank Environment")
    distribution: Optional[str] = Field(None, title="Distribution")
    time_limit: Optional[Uint32NoVal] = Field(None, title="Integer number with flags")
    time_minimum: Optional[Uint32NoVal] = Field(None, title="Integer number with flags")
    tres_bind: Optional[str] = Field(None, title="Tres Bind")
    tres_freq: Optional[str] = Field(None, title="Tres Freq")
    tres_per_job: Optional[str] = Field(None, title="Tres Per Job")
    tres_per_node: Optional[str] = Field(None, title="Tres Per Node")
    tres_per_socket: Optional[str] = Field(None, title="Tres Per Socket")
    tres_per_task: Optional[str] = Field(None, title="Tres Per Task")
    user_id: Optional[str] = Field(None, title="User Id")
    wait_all_nodes: Optional[bool] = Field(None, title="Wait All Nodes")
    kill_warning_flags: Optional[List[str]] = Field(None, title="Kill Warning Flags")
    kill_warning_signal: Optional[str] = Field(None, title="Kill Warning Signal")
    kill_warning_delay: Optional[Uint16NoVal] = Field(
        None, title="Integer number with flags"
    )
    current_working_directory: Optional[str] = Field(
        None, title="Current Working Directory"
    )
    cpus_per_task: Optional[int] = Field(None, title="Cpus Per Task")
    minimum_cpus: Optional[int] = Field(None, title="Minimum Cpus")
    maximum_cpus: Optional[int] = Field(None, title="Maximum Cpus")
    nodes: Optional[str] = Field(None, title="Nodes")
    minimum_nodes: Optional[int] = Field(None, title="Minimum Nodes")
    maximum_nodes: Optional[int] = Field(None, title="Maximum Nodes")
    minimum_boards_per_node: Optional[int] = Field(
        None, title="Minimum Boards Per Node"
    )
    minimum_sockets_per_board: Optional[int] = Field(
        None, title="Minimum Sockets Per Board"
    )
    sockets_per_node: Optional[int] = Field(None, title="Sockets Per Node")
    threads_per_core: Optional[int] = Field(None, title="Threads Per Core")
    tasks_per_node: Optional[int] = Field(None, title="Tasks Per Node")
    tasks_per_socket: Optional[int] = Field(None, title="Tasks Per Socket")
    tasks_per_core: Optional[int] = Field(None, title="Tasks Per Core")
    tasks_per_board: Optional[int] = Field(None, title="Tasks Per Board")
    ntasks_per_tres: Optional[int] = Field(None, title="Ntasks Per Tres")
    minimum_cpus_per_node: Optional[int] = Field(None, title="Minimum Cpus Per Node")
    memory_per_cpu: Optional[Uint64NoVal] = Field(
        None, title="Integer number with flags"
    )
    memory_per_node: Optional[Uint64NoVal] = Field(
        None, title="Integer number with flags"
    )
    temporary_disk_per_node: Optional[int] = Field(
        None, title="Temporary Disk Per Node"
    )
    selinux_context: Optional[str] = Field(None, title="Selinux Context")
    required_switches: Optional[Uint32NoVal] = Field(
        None, title="Integer number with flags"
    )
    standard_error: Optional[str] = Field(None, title="Standard Error")
    standard_input: Optional[str] = Field(None, title="Standard Input")
    standard_output: Optional[str] = Field(None, title="Standard Output")
    wait_for_switch: Optional[int] = Field(None, title="Wait For Switch")
    wckey: Optional[str] = Field(None, title="Wckey")
    x11: Optional[List[str]] = Field(None, title="X11")
    x11_magic_cookie: Optional[str] = Field(None, title="X11 Magic Cookie")
    x11_target_host: Optional[str] = Field(None, title="X11 Target Host")
    x11_target_port: Optional[int] = Field(None, title="X11 Target Port")


class JobInfo(BaseModel):
    account: Optional[str] = Field(None, title="Account")
    accrue_time: Optional[int] = Field(None, title="Accrue Time")
    admin_comment: Optional[str] = Field(None, title="Admin Comment")
    allocating_node: Optional[str] = Field(None, title="Allocating Node")
    array_job_id: Optional[Uint32NoVal] = Field(None, title="Integer number with flags")
    array_task_id: Optional[Uint32NoVal] = Field(
        None, title="Integer number with flags"
    )
    array_max_tasks: Optional[Uint32NoVal] = Field(
        None, title="Integer number with flags"
    )
    array_task_string: Optional[str] = Field(None, title="Array Task String")
    association_id: Optional[int] = Field(None, title="Association Id")
    batch_features: Optional[str] = Field(None, title="Batch Features")
    batch_flag: Optional[bool] = Field(None, title="Batch Flag")
    batch_host: Optional[str] = Field(None, title="Batch Host")
    flags: Optional[List[str]] = Field(None, title="Flags")
    burst_buffer: Optional[str] = Field(None, title="Burst Buffer")
    burst_buffer_state: Optional[str] = Field(None, title="Burst Buffer State")
    cluster: Optional[str] = Field(None, title="Cluster")
    cluster_features: Optional[str] = Field(None, title="Cluster Features")
    command: Optional[str] = Field(None, title="Command")
    comment: Optional[str] = Field(None, title="Comment")
    container: Optional[str] = Field(None, title="Container")
    container_id: Optional[str] = Field(None, title="Container Id")
    contiguous: Optional[bool] = Field(None, title="Contiguous")
    core_spec: Optional[int] = Field(None, title="Core Spec")
    thread_spec: Optional[int] = Field(None, title="Thread Spec")
    cores_per_socket: Optional[Uint16NoVal] = Field(
        None, title="Integer number with flags"
    )
    billable_tres: Optional[Float64NoVal] = Field(
        None, title="64 bit floating point number with flags"
    )
    cpus_per_task: Optional[Uint16NoVal] = Field(
        None, title="Integer number with flags"
    )
    cpu_frequency_minimum: Optional[Uint32NoVal] = Field(
        None, title="Integer number with flags"
    )
    cpu_frequency_maximum: Optional[Uint32NoVal] = Field(
        None, title="Integer number with flags"
    )
    cpu_frequency_governor: Optional[Uint32NoVal] = Field(
        None, title="Integer number with flags"
    )
    cpus_per_tres: Optional[str] = Field(None, title="Cpus Per Tres")
    cron: Optional[str] = Field(None, title="Cron")
    deadline: Optional[int] = Field(None, title="Deadline")
    delay_boot: Optional[Uint32NoVal] = Field(None, title="Integer number with flags")
    dependency: Optional[str] = Field(None, title="Dependency")
    derived_exit_code: Optional[Uint32NoVal] = Field(
        None, title="Integer number with flags"
    )
    eligible_time: Optional[int] = Field(None, title="Eligible Time")
    end_time: Optional[int] = Field(None, title="End Time")
    excluded_nodes: Optional[str] = Field(None, title="Excluded Nodes")
    exit_code: Optional[Uint32NoVal] = Field(None, title="Integer number with flags")
    extra: Optional[str] = Field(None, title="Extra")
    failed_node: Optional[str] = Field(None, title="Failed Node")
    features: Optional[str] = Field(None, title="Features")
    federation_origin: Optional[str] = Field(None, title="Federation Origin")
    federation_siblings_active: Optional[str] = Field(
        None, title="Federation Siblings Active"
    )
    federation_siblings_viable: Optional[str] = Field(
        None, title="Federation Siblings Viable"
    )
    gres_detail: Optional[List[str]] = Field(None, title="Gres Detail")
    group_id: Optional[int] = Field(None, title="Group Id")
    group_name: Optional[str] = Field(None, title="Group Name")
    het_job_id: Optional[Uint32NoVal] = Field(None, title="Integer number with flags")
    het_job_id_set: Optional[str] = Field(None, title="Het Job Id Set")
    het_job_offset: Optional[Uint32NoVal] = Field(
        None, title="Integer number with flags"
    )
    job_id: Optional[int] = Field(None, title="Job Id")
    job_resources: Optional[JobRes] = Field(None, title="Job res")
    job_size_str: Optional[List[str]] = Field(None, title="Job Size Str")
    job_state: Optional[str] = Field(None, title="Job State")
    last_sched_evaluation: Optional[int] = Field(None, title="Last Sched Evaluation")
    licenses: Optional[str] = Field(None, title="Licenses")
    mail_type: Optional[List[str]] = Field(None, title="Mail Type")
    mail_user: Optional[str] = Field(None, title="Mail User")
    max_cpus: Optional[Uint32NoVal] = Field(None, title="Integer number with flags")
    max_nodes: Optional[Uint32NoVal] = Field(None, title="Integer number with flags")
    mcs_label: Optional[str] = Field(None, title="Mcs Label")
    memory_per_tres: Optional[str] = Field(None, title="Memory Per Tres")
    name: Optional[str] = Field(None, title="Name")
    network: Optional[str] = Field(None, title="Network")
    nodes: Optional[str] = Field(None, title="Nodes")
    nice: Optional[int] = Field(None, title="Nice")
    tasks_per_core: Optional[Uint16NoVal] = Field(
        None, title="Integer number with flags"
    )
    tasks_per_tres: Optional[Uint16NoVal] = Field(
        None, title="Integer number with flags"
    )
    tasks_per_node: Optional[Uint16NoVal] = Field(
        None, title="Integer number with flags"
    )
    tasks_per_socket: Optional[Uint16NoVal] = Field(
        None, title="Integer number with flags"
    )
    tasks_per_board: Optional[Uint16NoVal] = Field(
        None, title="Integer number with flags"
    )
    cpus: Optional[Uint32NoVal] = Field(None, title="Integer number with flags")
    node_count: Optional[Uint32NoVal] = Field(None, title="Integer number with flags")
    tasks: Optional[Uint32NoVal] = Field(None, title="Integer number with flags")
    partition: Optional[str] = Field(None, title="Partition")
    prefer: Optional[str] = Field(None, title="Prefer")
    memory_per_cpu: Optional[Uint64NoVal] = Field(
        None, title="Integer number with flags"
    )
    memory_per_node: Optional[Uint64NoVal] = Field(
        None, title="Integer number with flags"
    )
    minimum_cpus_per_node: Optional[Uint16NoVal] = Field(
        None, title="Integer number with flags"
    )
    minimum_tmp_disk_per_node: Optional[Uint32NoVal] = Field(
        None, title="Integer number with flags"
    )
    power: Optional[JobInfoPower] = Field(None, title="Job info power")
    preempt_time: Optional[int] = Field(None, title="Preempt Time")
    preemptable_time: Optional[int] = Field(None, title="Preemptable Time")
    pre_sus_time: Optional[int] = Field(None, title="Pre Sus Time")
    hold: Optional[bool] = Field(None, title="Hold (true) or release (false) job")
    priority: Optional[Uint32NoVal] = Field(None, title="Request specific job priority")
    profile: Optional[List[str]] = Field(None, title="Profile")
    qos: Optional[str] = Field(None, title="Qos")
    reboot: Optional[bool] = Field(None, title="Reboot")
    required_nodes: Optional[str] = Field(None, title="Required Nodes")
    minimum_switches: Optional[int] = Field(None, title="Minimum Switches")
    requeue: Optional[bool] = Field(None, title="Requeue")
    resize_time: Optional[int] = Field(None, title="Resize Time")
    restart_cnt: Optional[int] = Field(None, title="Restart Cnt")
    resv_name: Optional[str] = Field(None, title="Resv Name")
    scheduled_nodes: Optional[str] = Field(None, title="Scheduled Nodes")
    selinux_context: Optional[str] = Field(None, title="Selinux Context")
    shared: Optional[List[str]] = Field(None, title="Shared")
    exclusive: Optional[List[str]] = Field(None, title="Exclusive")
    oversubscribe: Optional[bool] = Field(None, title="Oversubscribe")
    show_flags: Optional[List[str]] = Field(None, title="Show Flags")
    sockets_per_board: Optional[int] = Field(None, title="Sockets Per Board")
    sockets_per_node: Optional[Uint16NoVal] = Field(
        None, title="Integer number with flags"
    )
    start_time: Optional[int] = Field(None, title="Start Time")
    state_description: Optional[str] = Field(None, title="State Description")
    state_reason: Optional[str] = Field(None, title="State Reason")
    standard_error: Optional[str] = Field(None, title="Standard Error")
    standard_input: Optional[str] = Field(None, title="Standard Input")
    standard_output: Optional[str] = Field(None, title="Standard Output")
    submit_time: Optional[int] = Field(None, title="Submit Time")
    suspend_time: Optional[int] = Field(None, title="Suspend Time")
    system_comment: Optional[str] = Field(None, title="System Comment")
    time_limit: Optional[Uint32NoVal] = Field(None, title="Integer number with flags")
    time_minimum: Optional[Uint32NoVal] = Field(None, title="Integer number with flags")
    threads_per_core: Optional[Uint16NoVal] = Field(
        None, title="Integer number with flags"
    )
    tres_bind: Optional[str] = Field(None, title="Tres Bind")
    tres_freq: Optional[str] = Field(None, title="Tres Freq")
    tres_per_job: Optional[str] = Field(None, title="Tres Per Job")
    tres_per_node: Optional[str] = Field(None, title="Tres Per Node")
    tres_per_socket: Optional[str] = Field(None, title="Tres Per Socket")
    tres_per_task: Optional[str] = Field(None, title="Tres Per Task")
    tres_req_str: Optional[str] = Field(None, title="Tres Req Str")
    tres_alloc_str: Optional[str] = Field(None, title="Tres Alloc Str")
    user_id: Optional[int] = Field(None, title="User Id")
    user_name: Optional[str] = Field(None, title="User Name")
    maximum_switch_wait_time: Optional[int] = Field(
        None, title="Maximum Switch Wait Time"
    )
    wckey: Optional[str] = Field(None, title="Wckey")
    current_working_directory: Optional[str] = Field(
        None, title="Current Working Directory"
    )


class JobSubmission(BaseModel):
    script: Optional[str] = Field(
        None,
        title="Executable script (full contents) to run in batch step for all job components",
    )
    job: Optional[JobDescMsg] = Field(None, title="Job desc msg")
    jobs: Optional[List[JobDescMsg]] = Field(None, title="List of job desc msg")


class MetaSlurm(BaseModel):
    version: Optional[MetaSlurmVersion] = Field(None, title="Meta slurm version")
    release: Optional[str] = Field(None, title="version specifier")


class PostBodySlurmJobSubmit(BaseModel):
    content: Optional[JobSubmission] = Field(None, title="Job submission")


class Meta(BaseModel):
    plugin: Optional[MetaPlugin] = Field(None, title="Meta plugin")
    Slurm: Optional[MetaSlurm] = Field(None, title="Slurm information")


class Status(BaseModel):
    meta: Optional[Meta] = None
    errors: Optional[List[Error]] = Field(None, title="List of error")
    warnings: Optional[List[Warning]] = Field(None, title="List of warning")


class DeleteResponse200SlurmJobJobid(BaseModel):
    content: Status


class DeleteResponseDefaultSlurmJobJobid(BaseModel):
    content: Status


class GetResponseDefaultSlurmJobJobid(BaseModel):
    content: Status


class GetResponseDefaultSlurmJobs(BaseModel):
    content: Status


class JobSubmissionResponse(BaseModel):
    meta: Optional[Meta] = None
    errors: Optional[List[Error]] = Field(None, title="List of error")
    warnings: Optional[List[Warning]] = Field(None, title="List of warning")
    job_id: Optional[int] = Field(None, title="new job ID")
    step_id: Optional[str] = Field(None, title="new job step ID")
    job_submit_user_msg: Optional[str] = Field(
        None, title="Message to user from job_submit plugin"
    )


class JobsResponse(BaseModel):
    meta: Optional[Meta] = None
    errors: Optional[List[Error]] = Field(None, title="List of error")
    warnings: Optional[List[Warning]] = Field(None, title="List of warning")
    jobs: Optional[List[JobInfo]] = Field(None, title="List of job info")


class PostResponse200SlurmJobSubmit(BaseModel):
    content: JobSubmissionResponse = Field(..., title="Job submission response")


class PostResponseDefaultSlurmJobSubmit(BaseModel):
    content: Status


class GetResponse200SlurmJobJobid(BaseModel):
    content: JobsResponse = Field(..., title="Jobs response")


class GetResponse200SlurmJobs(BaseModel):
    content: JobsResponse = Field(..., title="Jobs response")


class AllModels(BaseModel):
    model0: PostPathSlurmJobSubmit = Field(
        ..., title="POST /slurm/v0.0.39/job/submit: path parameters"
    )
    model1: PostQuerySlurmJobSubmit = Field(
        ..., title="POST /slurm/v0.0.39/job/submit: query parameters"
    )
    model2: CronEntryLine = Field(..., title="Cron entry line")
    model3: CronEntry = Field(..., title="Cron entry")
    model4: Uint32NoVal = Field(..., title="Integer number with flags")
    model5: Uint16NoVal = Field(..., title="Integer number with flags")
    model6: Uint64NoVal = Field(..., title="Integer number with flags")
    model7: JobDescMsg = Field(..., title="Job desc msg")
    model8: JobSubmission = Field(..., title="Job submission")
    model9: PostBodySlurmJobSubmit = Field(..., title="submit new job")
    model10: MetaPlugin = Field(..., title="Meta plugin")
    model11: MetaSlurmVersion = Field(..., title="Meta slurm version")
    model12: MetaSlurm = Field(..., title="Slurm information")
    model13: Meta
    model14: Error
    model15: Warning
    model16: JobSubmissionResponse = Field(..., title="Job submission response")
    model17: PostResponse200SlurmJobSubmit = Field(..., title="job submitted")
    model18: Status
    model19: PostResponseDefaultSlurmJobSubmit = Field(..., title="job rejected")
    model20: GetPathSlurmJobJobid = Field(
        ..., title="GET /slurm/v0.0.39/job/{job_id}: path parameters"
    )
    model21: GetQuerySlurmJobJobid = Field(
        ..., title="GET /slurm/v0.0.39/job/{job_id}: query parameters"
    )
    model22: GetBodySlurmJobJobid = Field(
        ..., title="GET /slurm/v0.0.39/job/{job_id}: body"
    )
    model23: Float64NoVal = Field(..., title="64 bit floating point number with flags")
    model24: JobRes = Field(..., title="Job res")
    model25: JobInfoPower = Field(..., title="Job info power")
    model26: JobInfo = Field(..., title="Job info")
    model27: JobsResponse = Field(..., title="Jobs response")
    model28: GetResponse200SlurmJobJobid = Field(..., title="job(s) information")
    model29: GetResponseDefaultSlurmJobJobid = Field(
        ..., title="job matching JobId not found"
    )
    model30: DeletePathSlurmJobJobid = Field(
        ..., title="DELETE /slurm/v0.0.39/job/{job_id}: path parameters"
    )
    model31: DeleteQuerySlurmJobJobid = Field(
        ..., title="DELETE /slurm/v0.0.39/job/{job_id}: query parameters"
    )
    model32: DeleteBodySlurmJobJobid = Field(
        ..., title="DELETE /slurm/v0.0.39/job/{job_id}: body"
    )
    model33: DeleteResponse200SlurmJobJobid = Field(
        ..., title="job cancelled or sent signal"
    )
    model34: DeleteResponseDefaultSlurmJobJobid = Field(
        ..., title="Job cancel request failed"
    )
    model35: GetPathSlurmJobs = Field(
        ..., title="GET /slurm/v0.0.39/jobs: path parameters"
    )
    model36: GetQuerySlurmJobs = Field(
        ..., title="GET /slurm/v0.0.39/jobs: query parameters"
    )
    model37: GetBodySlurmJobs = Field(..., title="GET /slurm/v0.0.39/jobs: body")
    model38: GetResponse200SlurmJobs = Field(..., title="job(s) information")
    model39: GetResponseDefaultSlurmJobs = Field(..., title="job not found")



def add_parameters(parser):
    parser.add_argument(
        "job_ids",
        nargs="*",
        type=int,
        default=0,
        help="SLURM Job ID's to be cancelled",
    )
    parser.add_argument(
        "-a",
        "--all",
        action="store_true",
        help="All jobs when no job ID is provided",
    )


def apply_parameters(args):
    pass



import socket

JOB_NAME: str = f"pyslurmutils.{socket.gethostname()}"  # keep filename friendly
PYTHON_CMD: str = "python3"
SHEBANG: str = "#!/bin/bash -l"  # Assume Slurm is Linux
SLURM_TEMPDIR: str = "/tmp"  # Assume Slurm is Linux
SLURM_ARGUMENTS_NAME: str = "slurm_arguments"
MIN_PORT: int = 59000
NUM_PORTS: int = 100
DEFAULT_API_VERSION = "v0.0.41"



"""Remote execution: remote code declared on the client side with pure python builtins"""

from functools import lru_cache

try:
    from importlib.resources import files as resource_files
except ImportError:
    from importlib_resources import files as resource_files

# For testing: execute remote main locally
from ._tcp import main as tcp_main  # noqa F401
from ._file import main as file_main  # noqa F401


@lru_cache(3)
def remote_script(name: str) -> str:
    package, *parts = __name__.split(".")
    filename = resource_files(package)
    for part in parts:
        filename /= part
    filename /= f"_{name}.py"

    with open(filename, "r") as file:
        file_content = "".join(
            [s for s in file.readlines() if not s.startswith("from ._base")]
        )

    if name == "base":
        return file_content

    base = remote_script("base")
    return "".join(
        [
            base,
            "\n",
            file_content,
            "\n",
            "if __name__ == '__main__':\n",
            "    main()\n",
            "\n",
        ]
    )



import os
import getpass
import logging


_LOG_LEVELS = {
    "debug": logging.DEBUG,
    "info": logging.INFO,
    "warning": logging.WARNING,
    "error": logging.ERROR,
    "critical": logging.CRITICAL,
}


def add_parameters(parser):
    parser.add_argument(
        "-l",
        "--log",
        dest="log_level",
        type=str.lower,
        choices=list(_LOG_LEVELS),
        default="warning",
        help="Log level (default: %(default)s)",
    )
    parser.add_argument(
        "-u",
        "--user",
        dest="user_name",
        type=str,
        default=os.environ.get("SLURM_USER", getpass.getuser()),
        help="User name (default: %(default)s)",
    )
    parser.add_argument(
        "-t",
        "--token",
        type=str,
        default=os.environ.get("SLURM_TOKEN"),
        help="SLURM access token (default: %(default)s)",
    )
    parser.add_argument(
        "-v",
        "--api-version",
        type=str,
        default=os.environ.get("SLURM_API_VERSION"),
        help="SLURM access token (default: %(default)s)",
    )
    parser.add_argument(
        "--url",
        type=str,
        default=os.environ.get("SLURM_URL"),
        help="SLURM Rest Server (default: %(default)s)",
    )
    parser.add_argument(
        "--renewal-url",
        type=str,
        default=os.environ.get("SLURM_RENEWAL_URL"),
        help="URL for SLURM token renewal (default: %(default)s)",
    )
    if os.path.isdir(os.path.join(os.path.sep, "tmp_14_days")):
        log_directory = os.path.join(
            os.path.sep, "tmp_14_days", "{user_name}", "slurm_logs"
        )
    else:
        log_directory = None
    parser.add_argument(
        "--log-dir",
        dest="log_directory",
        type=str,
        default=log_directory,
        help="Directory of SLURM job logs (default: %(default)s)",
    )


def apply_parameters(args):
    logging.basicConfig(level=_LOG_LEVELS[args.log_level])



import os
import socket
import logging

from ._base import Client
from ._base import main_task_loop
from ._base import init_job


def main():
    init_job()
    with TcpClient() as client:
        main_task_loop(client)


class TcpClient(Client):
    def __init__(self) -> None:
        host = os.environ.get("_PYSLURMUTILS_HOST")
        port = int(os.environ.get("_PYSLURMUTILS_PORT"))
        try:
            hostname = socket.gethostbyaddr(host)[0]
        except Exception:
            hostname = host
        remote_name = f"{hostname}:{port}"

        self._client_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        logging.debug("Connecting to %s ...", remote_name)
        self._client_socket.settimeout(10)
        self._client_socket.connect((host, port))
        logging.debug("Connected to %s", remote_name)
        self._client_socket.settimeout(None)

        super().__init__(remote_name, remote_name)

    def close(self) -> None:
        logging.debug("Closing connection ...")
        try:
            self._client_socket.shutdown(socket.SHUT_WR)
        except Exception:
            pass
        finally:
            self._client_socket.close()
        logging.debug("Connection closed")

    def _send_bytes(self, bdata: bytes) -> None:
        self._client_socket.sendall(bdata)

    def _receive_nbytes(self, nbytes: int) -> bytes:
        data = bytearray()
        remaining = nbytes
        block = min(nbytes, self._MAX_CHUNK_NBYTES)

        while remaining > 0:
            chunk = self._client_socket.recv(min(block, remaining))
            if not chunk:
                raise ConnectionError(
                    f"Socket closed with {remaining} bytes left to read"
                )
            data.extend(chunk)
            remaining -= len(chunk)

        return bytes(data)



import os
from dataclasses import dataclass
from typing import List, Optional, Tuple

from pydantic import BaseModel


@dataclass
class StdInfo:
    """Information about a SLURM job stdout/stderr log file"""

    name: str
    filename: Optional[str]
    content: Optional[List[str]]

    def lines(self) -> List[str]:
        title = f"{self.name}: {self.filename}\n"
        line = "-" * len(title) + "\n"
        if self.content is None:
            return [title, line, " <not found>"]
        if not self.content:
            return [title, line, " <empty>"]
        return [title, line] + self.content


def read_slurm_log(
    job_properties: BaseModel, stderr: bool = False
) -> Tuple[Optional[str], Optional[List[str]]]:
    if stderr:
        filename = job_properties.standard_error
    else:
        filename = job_properties.standard_output
    filename = slurm_log_filename(job_properties, filename)
    if not filename:
        return None, None
    try:
        with open(filename) as f:
            return filename, list(f)
    except FileNotFoundError:
        return None, None


def remove_slurm_log(job_properties: BaseModel, stderr: bool = False) -> None:
    if stderr:
        filename = job_properties.standard_error
    else:
        filename = job_properties.standard_output
    filename = slurm_log_filename(job_properties, filename)
    if not filename or filename == "/dev/null":
        return
    try:
        os.remove(filename)
    except (FileNotFoundError, PermissionError):
        pass


def get_stdout_stderr(job_properties: BaseModel) -> Tuple[StdInfo, Optional[StdInfo]]:
    if (
        not job_properties.standard_error
        or job_properties.standard_output == job_properties.standard_error
    ):
        ofilename, ocontent = read_slurm_log(job_properties, stderr=False)
        std = StdInfo(name="STDOUT/STDERR", filename=ofilename, content=ocontent)
        return std, None

    ofilename, ocontent = read_slurm_log(job_properties, stderr=False)
    efilename, econtent = read_slurm_log(job_properties, stderr=True)
    stdout = StdInfo(name="STDOUT", filename=ofilename, content=ocontent)
    stderr = StdInfo(name="STDERR", filename=efilename, content=econtent)
    return stdout, stderr


def slurm_log_filename(
    job_properties: BaseModel, filename: Optional[str]
) -> Optional[str]:
    if filename:
        return filename.replace("%j", str(job_properties.job_id))



import os
import re
import abc
import sys
import inspect
import logging
from typing import Any, Optional, Tuple, Union

if sys.version_info >= (3, 8):
    from importlib.metadata import metadata as distribution_metadata
else:
    from importlib_metadata import metadata as distribution_metadata

from ... import errors


logger = logging.getLogger(__name__)

TASK = Union[callable, str]


class WorkerProxy:
    """Worker proxy to execute tasks with an optional maximum and an initializer.

    .. code-block::

        with WorkerProxy(initializer=..., initargs=..., initkwargs=..., max_tasks=...) as proxy:
            proxy.initialize()  # optionally explicit
            # -> max_tasks(int)
            # -> initializer(callable), initargs(tuple), initkwargs(dict), None
            # <- result, exception

            result1 = proxy.execute(task1, args1, kwargs1)
            # -> task1(callable), args1(tuple), kwargs1(dict)
            # <- result, exception

            result2 = proxy.execute(task2, args2, kwargs2)
            # -> task2(callable), args2(tuple), kwargs2(dict)
            # <- result, exception

            ...

            proxy.close()  # optionally explicit
            # -> None
    """

    def __init__(
        self,
        initializer: Optional[callable] = None,
        initargs: Optional[tuple] = None,
        initkwargs: Optional[tuple] = None,
        max_tasks: Optional[int] = None,
    ) -> None:
        """
        :param initializer: execute before the first task
        :param initargs: parameters for `initializer`
        :param initkwargs: parameters for `initializer`
        :param max_tasks: maximum number of tasks this proxy can execute
        """
        self._close_flag = False
        self._always_source = False

        self._initializer = initializer
        self._initargs = initargs
        self._initkwargs = initkwargs
        self._initial_send = False

        if max_tasks is not None and initializer is not None:
            # Initializers are send like any other task, only during initialization
            # and a failure will close the proxy.
            max_tasks += 1
        self._max_tasks = max_tasks
        self._task_counter = 0

    def close(self):
        if not self._close_flag:
            self._send_when_not_closed(None)
            self._close_flag = True

    def __enter__(self) -> "WorkerProxy":
        return self

    def __exit__(self, *_) -> None:
        self.close()

    def initialize(self) -> None:
        """Initialize the communication with the client. Will be done automatically upon the first `submit` call
        when not called explicitly. Re-raises the remote exception when the initializer fails.
        """
        if self._initial_send:
            return
        initial_data = self._max_tasks, logging.getLogger().getEffectiveLevel()
        self._send_when_not_closed(initial_data)
        self._initial_send = True
        if self._initializer is not None:
            try:
                _ = self.execute(self._initializer, self._initargs, self._initkwargs)
            except BaseException:
                self.close()
                raise

    def execute_without_reraise(
        self,
        task: Optional[callable],
        args: Optional[tuple] = None,
        kwargs: Optional[dict] = None,
    ) -> Tuple[Any, errors.SerializedException]:
        """Send the task to the client and wait for the result."""
        self.initialize()

        task_org = task
        if task is not None:
            task = _callable_or_source_code(task, always_source=self._always_source)

        self._send_when_not_closed((task, args, kwargs))
        self._task_counter += 1

        if task_org is None:
            logger.debug("stop request submitted to the remote worker")
        else:
            logger.debug("%s submitted to the remote worker", task_org)

        try:
            return self._get_result()
        finally:
            if self._max_tasks is not None and self._task_counter >= self._max_tasks:
                self.close()

    def execute(
        self,
        task: Optional[callable],
        args: Optional[tuple] = None,
        kwargs: Optional[dict] = None,
    ) -> Any:
        """Send the task to the client and wait for the result.
        When the task failed remotely, re-raise the exection locally.
        """
        result, exc_info = self.execute_without_reraise(task, args=args, kwargs=kwargs)
        if exc_info is not None:
            errors.reraise_remote_exception_from_tb(exc_info)
        return result

    def _send_when_not_closed(self, data: Any) -> None:
        if self._close_flag:
            raise RuntimeError("cannot send data after stopped")
        self._send(data)

    @abc.abstractmethod
    def _send(self, data: Any) -> None:
        pass

    @abc.abstractmethod
    def _get_result(self) -> Tuple[Any, Optional[errors.SerializedException]]:
        """In case of an exception, it could be raised (local error) or
        returned serialized (remote error)."""
        pass


def _callable_or_source_code(
    func: callable, always_source: bool = False
) -> Union[callable, Tuple[str, str]]:
    try:
        try:
            filename = inspect.getfile(func)
        except Exception:
            # For example a python builtin function
            return func
        if not filename or not os.path.isfile(filename):
            return func

        if not always_source:
            try:
                module = inspect.getmodule(func)
                package_name = module.__name__.split(".")[0]
                # TODO: distribution name and package name are not always the same
                _ = distribution_metadata(package_name)
            except Exception:
                pass
            else:
                if not module.__name__.startswith("pyslurmutils.tests"):
                    return func

        with open(filename, "r") as file:
            source_code = []
            for line in file.readlines():
                source_code.append(_safe_import(line))

        return func.__name__, "".join(source_code)
    except Exception:
        return func


def _safe_import(line: str) -> str:
    original_line = line

    line = line.rstrip()
    if line.startswith("import "):
        # Example: import a, b as bb
        imports = line[len("import ") :].split(",")
        try_block = f"    {line}"
        except_lines = []
        for imp in imports:
            imp = imp.strip()
            if " as " in imp:
                mod, alias = map(str.strip, imp.split(" as "))
                except_lines.append(f"    {alias} = None")
            else:
                except_lines.append(f"    {imp} = None")
        except_block = "\n".join(except_lines)
        return f"try:\n{try_block}\nexcept ImportError:\n{except_block}\n"

    if line.startswith("from "):
        # Example: from x.y import a as aa, b
        m = re.match(r"from\s+([^\s]+)\s+import\s+(.+)", line)
        if not m:
            return line
        module, names = m.groups()
        try_block = f"    {line}"
        except_lines = []
        for name in names.split(","):
            name = name.strip()
            if " as " in name:
                orig, alias = map(str.strip, name.split(" as "))
                except_lines.append(f"    {alias} = None")
            else:
                except_lines.append(f"    {name} = None")
        except_block = "\n".join(except_lines)
        return f"try:\n{try_block}\nexcept ImportError:\n{except_block}\n"

    return original_line
